import logging
import random
import requests
import six
import uuid
from datetime import timedelta
from django import forms
from django.core.urlresolvers import reverse
from django.contrib.auth import authenticate
from django.db.models import Q
from django.http import JsonResponse
from django.utils import timezone
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import View
from ipware.utils import is_valid_ipv4, is_valid_ipv6
from threading import Thread
from warmama import models, exceptions, fields
from warmama.conf import settings
from warmama import decorators
from warmama.serializers import MatchSerializer

logger = logging.getLogger(__name__)


##########
# Thread Workers - TODO move to separate file?
##########


def _remote_login(data):
    """Pass the login request to a remote url

    This asks the remote server to validate a login/password combination. It
    will send the results by calling the callback_url given.
    """
    # TODO - rewrite this and skip `requests`, its a shame to pull it in for this one function

    logger.info('Remote login sending request login: %s, handle: %d', data['login'], data['handle'])
    try:
        r = requests.post(settings.WARMAMA_AUTH_URL, timeout=5, data=data)
    except requests.ConnectionError as e:
        # TODO cleanup login session
        logger.info('Remote login request failed: %s', e)
        return
    except requests.Timeout:
        # TODO retry or cleanup login session
        logger.info('Remote login request timeout login: %s, handle: %d', data['login'], data['handle'])
        return

    logger.info('Remote login response login: %s, handle: %d, response: %s', data['login'], data['handle'], r.text)


##########
# Helpers
##########


def _ipv4_ipv6(ipaddr):
    """Determine if an address is ipv4 or ipv6

    Returns:
        The tuple `(ipv4, ipv6)` where one is the string address and the other
        is an empty string.
    """
    # TODO this is repeated in warmama.decorators, move this to a utils file?
    if is_valid_ipv4(ipaddr):
        return ipaddr, ''
    elif is_valid_ipv6(ipaddr):
        return '', ipaddr

    raise exceptions.BadRequest('Couldnt read IP address for request')


def _ip_flt(ipv4, ipv6):
    """Queryset filter to match either ipv4 or ipv6"""
    assert ipv4 or ipv6, 'One of ipv4, ipv6 must be defined'
    return Q(ip=ipv4) | Q(ipv6=ipv6)


##########
# Base View Classes
##########


class APIMeta(type):
    """Metaclass to transform the `post` method of a view.

    The `post` method is trasformed in the following way:

        - The keyword arguments `POST`, `ipv4`, and `ipv6` are added to the
          method
        - The argument `POST` is the `request.POST` data cleaned and validated
          using the form `View.PostForm`
        - The arguments `ipv4` and `ipv6` is the source ip address for the
          request. It is guaranteed that exactly one of these is a valid address
          and the other is an emtpy string.
        - View exceptions are handled by `warmama.decorators.exception_response`
          and return `View.exc_response` on errors.

    This mainly prevents having to stack infinity decorators on each post method
    """
    def __new__(cls, name, parents, attrs):
        post = attrs.pop('post')
        klass = super(APIMeta, cls).__new__(cls, name, parents, attrs)

        # order matters
        post = decorators.log_args(logger)(post)
        post = method_decorator(decorators.request_ip)(post)
        post = method_decorator(decorators.clean_post_args(klass.PostForm))(post)
        post = decorators.exception_response(klass.exc_response)(post)
        klass.post = post

        return klass


class APIView(six.with_metaclass(APIMeta, View)):
    """Base class for an API endpoint

    Offer benefits from APIMeta and disables CSRF protection.
    """
    exc_response = {'status': 0}

    class PostForm(forms.Form):
        pass

    @classmethod
    def as_view(cls, *args, **kwargs):
        """Disable csrf protection"""
        view = super(APIView, cls).as_view(*args, **kwargs)
        return csrf_exempt(view)

    def post(self, request, *args, **kwargs):
        raise NotImplemented('APIViews must implement a post method')


##########
# View Classes
##########


class ServerLogin(APIView):
    """Server Login

    This allows for one server to have multiple sessions, one session per port.
    """
    exc_response = {'id': 0}

    class PostForm(forms.Form):
        port = forms.IntegerField()
        authkey = forms.CharField(max_length=64)
        hostname = forms.CharField(max_length=64, required=False)
        demos_baseurl = forms.CharField(max_length=128, required=False)

    def post(self, request, POST=None, ipv4='', ipv6=''):
        server = models.Server.objects.get(
            _ip_flt(ipv4, ipv6),
            login=POST['authkey'],
            banned=False,
        )

        server.hostname = POST['hostname']
        server.demos_baseurl = POST['demos_baseurl']
        server.save()

        session, created = models.ServerSession.objects.get_or_create(
            user=server,
            ip=ipv4,
            ipv6=ipv6,
            defaults={'port': POST['port']},
        )

        if not created:
            session.port = POST['port']
            session.save()
            # TODO: reset state (kick players, purge player, etc...)

        return JsonResponse({'id': session.pk})


class ServerLogout(APIView):
    """Server Logout"""
    class PostForm(forms.Form):
        ssession = forms.IntegerField(min_value=1)

    def post(self, request, POST=None, ipv4='', ipv6=''):
        """
        TODO: First remove all purge_players that match this server for all
        players who's server session matches reset their servers or if their
        purgable flag is set, remove them totally
        """
        session = models.ServerSession.objects.get(
            _ip_flt(ipv4, ipv6),
            pk=POST['ssession'],
        )
        logger.info('%s:%d logging off', ipv4 or ipv6, session.port)
        session.delete()
        return JsonResponse({'status': 1})


class ServerClientConnect(APIView):
    """Server client connect"""
    exc_response = {'id': 0}

    class PostForm(forms.Form):
        ssession = forms.IntegerField(min_value=1)
        csession = forms.IntegerField(min_value=1)
        cticket = forms.IntegerField(min_value=1)
        cip = fields.IPAddressField()

    def post(self, request, POST=None, ipv4='', ipv6=''):
        server_session = models.ServerSession.objects.get(
            _ip_flt(ipv4, ipv6),
            pk=POST['ssession'],
        )

        caddr, _ = POST['cip']
        cipv4, cipv6 = _ipv4_ipv6(caddr)
        logger.info(
            'SCC pk %d, tid %d, ts %d, now %s',
            POST['csession'], POST['cticket'], server_session.user_id, timezone.now()
        )
        client_session = models.PlayerSession.objects \
            .select_related('user') \
            .get(
                _ip_flt(cipv4, cipv6),
                pk=POST['csession'],
                ticket_id=POST['cticket'],
                ticket_server=server_session,
                ticket_expiration__gt=timezone.now(),
            )

        logger.info(
            'Attaching client session %s to server session %s',
            client_session, server_session
        )
        client_session.ticket_id = None
        client_session.ticket_server = None
        client_session.ticket_expiration = None
        client_session.server_session = server_session
        client_session.save()

        stats = models.PlayerStat.objects \
                                 .select_related('gametype') \
                                 .filter(player_id=client_session.user_id) \
                                 .order_by('gametype__name')

        data = {
            'id': client_session.pk,
            'login': client_session.user.login,
            'ratings': [{
                'gametype': stat.gametype.name,
                'rating': float(stat.rating),
                'deviation': float(stat.deviation),
            } for stat in stats],
        }
        return JsonResponse(data)


class ServerClientDisconnect(APIView):
    """Server client disconnect"""
    class PostForm(forms.Form):
        ssession = forms.IntegerField(min_value=1)
        csession = forms.IntegerField(min_value=1)
        gameon = forms.IntegerField(required=False)

    def post(self, request, POST=None, ipv4='', ipv6=''):
        server_session = models.ServerSession.objects.get(
            _ip_flt(ipv4, ipv6),
            pk=POST['ssession'],
        )
        client_session = models.PlayerSession.objects.get(
            pk=POST['csession'],
            server_session=server_session,
        )

        if POST['gameon']:
            logger.info('Adding session %d to purgables', client_session.pk)
            models.PurgePlayer.objects.create(
                player_id=client_session.user_id,
                session=client_session,
                server_session=server_session,
            )

        client_session.ticket_id = None
        client_session.ticket_server = None
        client_session.ticket_expiration = None
        client_session.server_session = None
        client_session.save()

        logger.info(
            'ServerClientDisconnect %d %d ok',
            client_session.pk, server_session.pk
        )
        return JsonResponse({'status': 1})


class ServerHeartbeat(APIView):
    """Server heartbeat"""
    class PostForm(forms.Form):
        ssession = forms.IntegerField(min_value=1)

    def post(self, request, POST=None, ipv4='', ipv6=''):
        server_session = models.ServerSession.objects.get(
            _ip_flt(ipv4, ipv6),
            pk=POST['ssession'],
        )

        # this will set the `updated` time
        server_session.save()

        logger.info('Heartbeat from server: %s', ipv4 or ipv6)
        return JsonResponse({'status': 1})


class ServerMatchReport(APIView):
    """Server match report"""
    class PostForm(forms.Form):
        ssession = forms.IntegerField()
        data = fields.GzipJsonField()

    def post(self, request, POST=None, ipv4='', ipv6=''):
        server_session = models.ServerSession.objects.get(
            _ip_flt(ipv4, ipv6),
            pk=POST['ssession'],
        )

        serializer = MatchSerializer(data=POST['data'])
        if not serializer.is_valid():
            raise exceptions.BadRequest(serializer.errors)

        matchresult, matchplayers = serializer.save(server_session=server_session)
        ratings = {
            'gametype': matchresult.gametype.name,
            'ratings': [
                [sessionid, float(matchplayer.newrating)]
                for sessionid, matchplayer in matchplayers.items()
            ]
        }

        # delete purgables, this cascade deletes PurgePlayer objects
        models.PlayerSession.objects \
                            .filter(server_session=server_session, purgable=True) \
                            .delete()

        # TODO: clear server_session.next_match_uuid?

        return JsonResponse({'status': 1, 'ratings': ratings})


class ServerMatchUUID(APIView):
    """Server Match UUID"""
    exc_response = {'uuid': ''}

    class PostForm(forms.Form):
        ssession = forms.IntegerField()

    def post(self, request, POST=None, ipv4='', ipv6=''):
        server_session = models.ServerSession.objects.get(
            _ip_flt(ipv4, ipv6),
            pk=POST['ssession'],
        )

        # TODO: collision check - probably not necessary
        # also, why do we even need to do this here. Best I can tell, gameserver
        # sets this to a configstring which is never actually used.
        server_session.next_match_uuid = str(uuid.uuid4())
        server_session.save()
        logger.info(
            'MatchUUID request from %s, returning %s',
            ipv4 or ipv6, server_session.next_match_uuid
        )
        return JsonResponse({'uuid': server_session.next_match_uuid})


class ClientLogin(APIView):
    """Client Login"""
    exc_response = {'ready': 2, 'id': 0}

    class PostForm(forms.Form):
        login = forms.CharField(max_length=64, required=False)
        passwd = forms.CharField(max_length=64, required=False)
        handle = forms.IntegerField(required=False)

    def _step_one(self, request, POST=None, ipv4='', ipv6=''):
        """Start the login process"""
        logger.info(
            'Clientlogin 1 login: %s, ip: %s',
            POST['login'], ipv4 or ipv6
        )

        # Create LoginPlayer for the login process, handle is pk for this
        # If one already exists for the login, let it error
        login_player = models.LoginPlayer.objects.create(login=POST['login'], ready=False, valid=False)

        if settings.WARMAMA_REMOTE_CLIENTAUTH:
            # Use remote authentication
            # TODO - digest should be a unique identifier to associate the
            # callbacks with this request, need to move `digest` field from
            # PlayerSession to LoginPlayer to do that
            thread_args = ({
                'login': POST['login'],
                'passwd': POST['passwd'],
                'handle': '%d' % login_player.pk,
                'digest': 'something',
                'url': request.build_absolute_uri(reverse('warmama:auth', current_app=request.resolver_match.namespace)),
            },)
            Thread(target=_remote_login, args=thread_args).start()
        else:
            # Use local authentication
            user = authenticate(username=POST['login'], password=POST['passwd'])
            login_player.ready = True
            login_player.valid = bool(user)
            login_player.save()

        return JsonResponse({'ready': -1, 'handle': login_player.pk, 'id': 0})

    def _step_two(self, request, POST=None, ipv4='', ipv6=''):
        """Complete the login process"""
        logger.info(
            'Clientlogin 2 handle: %d, ip: %s',
            POST['handle'], ipv4 or ipv6
        )

        # TODO - ip validation should be done here, but that means modifying
        # the DB schema
        login_player = models.LoginPlayer.objects.get(pk=POST['handle'])

        if not login_player.ready:
            logger.info(
                'Clientlogin 2 not ready. handle: %d, ip: %s',
                POST['handle'], ipv4 or ipv6
            )
            return JsonResponse({'ready': 1, 'id': 0})

        login_player.delete()

        if not login_player.valid:
            raise exceptions.Forbidden('User failed to authenticate')

        # Valid login, make a player session
        # TODO add session digest, still don't know what its used for
        player = models.Player.objects.get(login=login_player.login)
        player_session, created = models.PlayerSession.objects.get_or_create(user=player)

        if not created:
            if player_session.server_session_id:
                raise exceptions.Forbidden('User on server for existing session')

            if player_session.ip != ipv4 and player_session.ipv6 != ipv6:
                raise exceptions.Forbidden('Different IP for existing session')

        player_session.ip = ipv4
        player_session.ipv6 = ipv6
        player_session.purgable = False
        player_session.server_session = None
        player_session.save()

        # Calculate player profile
        stats = models.PlayerStat.objects \
                                 .select_related('gametype') \
                                 .filter(player=player) \
                                 .order_by('gametype__name')
        urlargs = {'session': player_session.pk}

        data = {
            'ready': 2,
            'id': player_session.pk,
            'ratings': [{
                'gametype': stat.gametype.name,
                'rating': float(stat.rating),
                'deviation': float(stat.deviation),
            } for stat in stats],
            'profile_url': login_player.profile_url.format(**urlargs),
            'profile_url_rml': login_player.profile_url_rml.format(**urlargs),
        }
        return JsonResponse(data)

    def post(self, request, POST=None, ipv4='', ipv6=''):
        """
        Returns {'ready': value, 'id': value} for one of the following cases:
            `{'ready': -1, 'handle': handle, 'id': 0}` - handle for the login process
            `{'ready': 1, 'id': 0} - login session isn't ready
            `{'ready': 2, 'id': sessionid } - login session is ready with id `id`
        """
        if POST.get('handle', None):
            return self._step_two(request, POST, ipv4, ipv6)
        else:
            if not POST.get('login', '') or not POST.get('passwd', ''):
                raise exceptions.BadRequest('Login and passwd fields are required')
            return self._step_one(request, POST, ipv4, ipv6)


class ClientLogout(APIView):
    """Client Logout"""
    class PostForm(forms.Form):
        csession = forms.IntegerField(min_value=1)

    def post(self, request, POST=None, ipv4='', ipv6=''):
        client_session = models.PlayerSession.objects.get(
            _ip_flt(ipv4, ipv6),
            pk=POST['csession'],
        )

        logger.info('Client %s logging off', ipv4 or ipv6)

        if hasattr(client_session, 'purge_player'):
            logger.info('Marking session %d purgable', client_session.pk)
            client_session.purgable = True
            client_session.ticket_id = None
            client_session.ticket_server = None
            client_session.ticket_expiration = None
            client_session.server_session = None
            client_session.save()
        else:
            logger.info('Removing session %d', client_session.pk)
            client_session.delete()

        return JsonResponse({'status': 1})


class ClientConnect(APIView):
    """Client Connect"""
    exc_response = {'ticket': 0}

    class PostForm(forms.Form):
        csession = forms.IntegerField(min_value=1)
        saddr = fields.IPAddressField()

    def post(self, request, POST=None, ipv4='', ipv6=''):
        sipv4, sipv6 = _ipv4_ipv6(POST['saddr'][0])  # TODO
        port = POST['saddr'][1] or settings.WARMAMA_DEFAULT_SERVER_PORT

        server_session = models.ServerSession.objects.get(
            _ip_flt(sipv4, sipv6),
            port=port
        )

        client_session = models.PlayerSession.objects.get(
            _ip_flt(ipv4, ipv6),
            pk=POST['csession'],
        )

        logger.info(
            'Found server %d for client %d (addr %s)',
            server_session.pk, client_session.pk, sipv4 or sipv6
        )

        client_session.ticket_id = random.randint(1, 0xfffffff)
        client_session.ticket_server = server_session
        client_session.ticket_expiration = timezone.now() + timedelta(seconds=settings.WARMAMA_TICKET_EXPIRATION)
        client_session.server_session = None
        client_session.save()

        return JsonResponse({'ticket': client_session.ticket_id})


class ClientHeartbeat(APIView):
    """Client Heartbeat"""
    class PostForm(forms.Form):
        csession = forms.IntegerField(min_value=1)

    def post(self, request, POST=None, ipv4='', ipv6=''):
        client_session = models.PlayerSession.objects.get(
            _ip_flt(ipv4, ipv6),
            pk=POST['csession'],
        )

        # this will set the `updated` time
        client_session.save()

        logger.info('Heartbeat from client: %s', ipv4 or ipv6)
        return JsonResponse({'status': 1})


class ClientAuthenticate(APIView):
    """Client Authenticate

    This is the callback endpoint for remote client auth requests
    """
    class PostForm(forms.Form):
        handle = forms.IntegerField(min_value=1)
        digest = forms.CharField(max_length=64)
        valid = forms.BooleanField()
        profile_url = forms.CharField(max_length=255, required=False)
        profile_url_rml = forms.CharField(max_length=255, required=False)

    def post(self, request, POST=None, ipv4='', ipv6=''):
        # TODO validate this with digest

        handle = POST.pop('handle')
        POST.pop('digest')

        models.LoginPlayer.objects.filter(pk=handle).update(**POST)

        return JsonResponse({'status': 1})



import json
import mock
import time
import zlib
from base64 import urlsafe_b64encode
from datetime import timedelta
from django.contrib.auth import get_user_model
from django.test import Client, override_settings
from django.utils import timezone
from tests import TestCase, loadJson
from warmama import models


##########
# Base classes for test cases
##########


class SampleData(object):
    """Mixin to setup sample data for test cases"""

    def setUp(self):
        """Set up data and test client for tests"""
        super(SampleData, self).setUp()
        User = get_user_model()
        self.user = User.objects.create_user('user', 'user@email.com', 'pass')

        self.server = models.Server.objects.create(
            login='server',
            ipv6='1:2:3:4:5:6:7:8',
        )
        self.ssession = models.ServerSession.objects.create(
            user=self.server,
            ipv6=self.server.ipv6,
            port=44400,
        )
        self.player = models.Player.objects.create(
            login='user',
            ip='132.233.139.177',
        )
        self.csession = models.PlayerSession.objects.create(
            user=self.player,
            ip=self.player.ip,
            ticket_id=1,
            ticket_server=self.ssession,
            ticket_expiration=timezone.now() + timedelta(days=1),
        )
        self.gametypes = [
            models.Gametype.objects.create(
                name='gt_{0}'.format(name),
                description='Gametype {0}'.format(name),
            ) for name in ('gt_b', 'gt_a')
        ]
        self.pstats = [
            models.PlayerStat.objects.create(
                player_id=self.player.pk,
                gametype_id=self.gametypes[0].pk,
                rating=4,
            ),
            models.PlayerStat.objects.create(
                player_id=self.player.pk,
                gametype_id=self.gametypes[1].pk,
                rating=5,
            )
        ]


class ServerTestCase(SampleData, TestCase):
    def setUp(self):
        """Attach a test client with server ip"""
        super(ServerTestCase, self).setUp()
        self.c = Client(REMOTE_ADDR='1:2:3:4:5:6:7:8')


class ClientTestCase(SampleData, TestCase):
    def setUp(self):
        """Attach a test client with client ip"""
        super(ClientTestCase, self).setUp()
        self.c = Client(REMOTE_ADDR='132.233.139.177')


##########
# Test Cases
##########


class ServerLoginTest(ServerTestCase):
    def test_unregistered_server(self):
        """It should return 403 on unknown ip addresses"""
        c = Client(REMOTE_ADDR='253.168.132.100')
        response = c.post('/slogin', {'authkey': 'login', 'port': 44400})

        self.assertEqual(response.status_code, 403)
        self.assertJSONEqual(response.content, {'id': 0})

    def test_invalid_authkey(self):
        """It should return 403 on unknown ip addresses"""
        response = self.c.post('/slogin', {'authkey': 'invalid', 'port': 44400})

        self.assertEqual(response.status_code, 403)
        self.assertJSONEqual(response.content, {'id': 0})

    def test_banned_server(self):
        """It should return 403 on banned servers"""
        self.server.banned = True
        self.server.save()

        response = self.c.post('/slogin', {'authkey': 'login', 'port': 44400})
        self.assertEqual(response.status_code, 403)
        self.assertJSONEqual(response.content, {'id': 0})

    def test_update_server(self):
        """It should update server hostname and demos_baseurl"""
        response = self.c.post('/slogin', {
            'authkey': 'server',
            'port': 44400,
            'hostname': 'hostname',
            'demos_baseurl': 'baseurl',
        })
        server = models.Server.objects.get(pk=self.server.pk)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(server.hostname, 'hostname')
        self.assertEqual(server.demos_baseurl, 'baseurl')

    def test_create_session(self):
        """It should create a session if there is none"""
        self.ssession.delete()
        response = self.c.post('/slogin', {'authkey': 'server', 'port': 44400})
        session = models.ServerSession.objects.get(user_id=self.server.pk)

        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(response.content, {'id': session.pk})

    def test_existing_session(self):
        """It should return an existing session if possible"""
        response = self.c.post('/slogin', {'authkey': 'server', 'port': 44400})

        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(response.content, {'id': self.ssession.pk})

    def test_update_session(self):
        """It should update an existing session's port"""
        response = self.c.post('/slogin', {
            'authkey': 'server',
            'port': '1234'
        })
        session = models.ServerSession.objects.get(pk=self.ssession.pk)

        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(response.content, {'id': session.pk})
        self.assertEqual(session.port, 1234)

    def test_invalid_port(self):
        """It should 400 on an invalid session port"""
        response = self.c.post('/slogin', {
            'authkey': 'login',
            'port': 'abcd'
        })

        self.assertEqual(response.status_code, 400)
        self.assertJSONEqual(response.content, {'id': 0})


class ServerLogoutTest(ServerTestCase):
    def test_invalid_ipaddr(self):
        """It should 403 on an invalid ip address"""
        c = Client(REMOTE_ADDR='132.233.139.177')
        response = c.post('/slogout', {'ssession': self.ssession.pk})

        self.assertEqual(response.status_code, 403)
        self.assertJSONEqual(response.content, {'status': 0})

    def test_invalid_session(self):
        """It should 403 on an invalid session"""
        response = self.c.post('/slogout', {'ssession': 1234})

        self.assertEqual(response.status_code, 403)
        self.assertJSONEqual(response.content, {'status': 0})

    def test_logout(self):
        """It should delete an existing session"""
        response = self.c.post('/slogout', {'ssession': self.ssession.pk})

        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(response.content, {'status': 1})
        with self.assertRaises(models.ServerSession.DoesNotExist):
            models.ServerSession.objects.get(pk=self.ssession.pk)


class ServerClientConnectTest(ServerTestCase):
    def test_expired_ticket(self):
        """It should 403 on expired tickets"""
        self.csession.ticket_expiration = timezone.now() - timedelta(days=1)
        self.csession.save()
        response = self.c.post('/scc', {
            'ssession': self.ssession.pk,
            'csession': self.csession.pk,
            'cticket': self.csession.ticket_id,
            'cip': self.csession.ip + ':1234'
        })

        self.assertEqual(response.status_code, 403)
        self.assertJSONEqual(response.content, {'id': 0})

    def test_reset_ticket(self):
        """It should reset the player's ticket"""
        response = self.c.post('/scc', {
            'ssession': self.ssession.pk,
            'csession': self.csession.pk,
            'cticket': self.csession.ticket_id,
            'cip': self.csession.ip + ':1234'
        })
        csession = models.PlayerSession.objects.get(pk=self.csession.pk)

        self.assertEqual(response.status_code, 200)
        self.assertIsNone(csession.ticket_id)
        self.assertIsNone(csession.ticket_server)
        self.assertIsNone(csession.ticket_expiration)

    def test_empty_ratings(self):
        """It should handle empty player ratings"""
        models.PlayerStat.objects.all().delete()
        response = self.c.post('/scc', {
            'ssession': self.ssession.pk,
            'csession': self.csession.pk,
            'cticket': self.csession.ticket_id,
            'cip': self.csession.ip + ':1234'
        })

        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(response.content, {
            'id': self.csession.pk,
            'login': self.player.login,
            'ratings': [],
        })

    def test_ratings(self):
        """It should handle player ratings"""
        response = self.c.post('/scc', {
            'ssession': self.ssession.pk,
            'csession': self.csession.pk,
            'cticket': self.csession.ticket_id,
            'cip': self.csession.ip + ':1234'
        })
        resdata = loadJson(response.content)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(resdata['id'], self.csession.pk)
        self.assertEqual(resdata['login'], self.player.login)
        self.assertEqual(
            resdata['ratings'],
            [{  # Should be ordered by gametype name
                'gametype': self.gametypes[1].name,
                'rating': 5.0,
                'deviation': 0,
            }, {
                'gametype': self.gametypes[0].name,
                'rating': 4.0,
                'deviation': 0,
            }]
        )


class ServerClientDisconnectTest(ServerTestCase):
    def setUp(self):
        """Connect client session to server session"""
        super(ServerClientDisconnectTest, self).setUp()
        self.csession.ticket_id = None
        self.csession.ticket_server = None
        self.csession.ticket_expiration = None
        self.csession.server_session = self.ssession
        self.csession.save()

    def test_client_session_permission(self):
        """It should 403 if client session does not belong to server session"""
        self.csession.server_session = None
        self.csession.save()
        response = self.c.post('/scd', {
            'ssession': self.ssession.pk,
            'csession': self.csession.pk,
        })

        self.assertEqual(response.status_code, 403)
        self.assertJSONEqual(response.content, {'status': 0})

    def test_client_disconnect(self):
        """It should clear ticket and session and not add a PlayerPurge"""
        response = self.c.post('/scd', {
            'ssession': self.ssession.pk,
            'csession': self.csession.pk,
        })
        csession = models.PlayerSession.objects.get(pk=self.csession.pk)

        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(response.content, {'status': 1})
        self.assertIsNone(csession.ticket_id)
        self.assertIsNone(csession.server_session)

    def test_nopurge(self):
        """It does not purge if `gameon` is 0 or not passed"""
        self.c.post('/scd', {
            'ssession': self.ssession.pk,
            'csession': self.csession.pk,
            'gameon': '0',
        })

        with self.assertRaises(models.PurgePlayer.DoesNotExist):
            models.PurgePlayer.objects.get(session=self.csession)

    def test_purge(self):
        """It add a PurgePlayer if `gameon` is passed"""
        self.c.post('/scd', {
            'ssession': self.ssession.pk,
            'csession': self.csession.pk,
            'gameon': 1,
        })
        purge = models.PurgePlayer.objects.get(
            session_id=self.csession.pk,
            server_session=self.ssession.pk,
        )

        self.assertIsNotNone(purge)


class ServerHeartbeat(ServerTestCase):
    def test_heartbeat(self):
        """It should update the `updated` timestamp"""
        # Force set `updated` to a time in the past and make sure it worked
        models.ServerSession.objects.update(updated=timezone.now() - timedelta(days=1))
        ssession = models.ServerSession.objects.get(pk=self.ssession.pk)
        self.assertTrue(timezone.now() - ssession.updated > timedelta(minutes=1))

        response = self.c.post('/shb', {'ssession': self.ssession.pk})
        ssession = models.ServerSession.objects.get(pk=self.ssession.pk)

        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(response.content, {'status': 1})
        self.assertTrue(timezone.now() - ssession.updated < timedelta(minutes=1))


class ServerMatchReportTest(ServerTestCase):
    def test_matchreport(self):
        data = {
            'match': {
                'gametype': 'ca',
                'teamgame': False,
                'racegame': False,
                'map': 'wca1',
                'hostname': 'host',
                'timeplayed': 100,
                'timestamp': 0,
                'gamedir': 'basewsw',
            },
            'players': [{
                'final': 1,
                'frags': 0,
                'deaths': 2,
                'sessionid': self.csession.pk,
            }],
        }
        packed = json.dumps(data).encode('utf-8')
        packed = zlib.compress(packed)
        packed = urlsafe_b64encode(packed).decode('ascii')
        expected = {
            'status': 1,
            'ratings': {
                'gametype': 'ca',
                'ratings': [[self.csession.pk, 0]],
            }
        }

        response = self.c.post('/smr', {'ssession': self.ssession.pk, 'data': packed})
        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(response.content, expected)


class ServerMatchUUIDTest(ServerTestCase):
    def test_matchuuid(self):
        response = self.c.post('/smuuid', {'ssession': self.ssession.pk})
        content = json.loads(response.content.decode('utf-8'))
        ssession = models.ServerSession.objects.get(pk=self.ssession.pk)

        self.assertEqual(response.status_code, 200)
        self.assertTrue(content['uuid'])
        self.assertEqual(ssession.next_match_uuid, content['uuid'])


class ClientLoginTest(ClientTestCase):
    @override_settings(WARMAMA_REMOTE_CLIENTAUTH=False)
    @mock.patch('requests.post')
    def test_step_one_local(self, requests_post):
        """It should mark the login_player as valid"""
        response = self.c.post('/clogin', {'login': 'user', 'passwd': 'pass'})
        login_player = models.LoginPlayer.objects.get(login='user')

        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(response.content, {
            'ready': -1,
            'handle': login_player.pk,
            'id': 0,
        })
        self.assertTrue(login_player.ready)
        self.assertTrue(login_player.valid)
        self.assertFalse(requests_post.called)

    @override_settings(WARMAMA_REMOTE_CLIENTAUTH=False)
    def test_step_one_invalid_user(self):
        """It should mark the login_player as invalid"""
        response = self.c.post('/clogin', {'login': 'user', 'passwd': 'wrong'})
        login_player = models.LoginPlayer.objects.get(login='user')

        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(response.content, {
            'ready': -1,
            'handle': login_player.pk,
            'id': 0,
        })
        self.assertTrue(login_player.ready)
        self.assertFalse(login_player.valid)

    @override_settings(WARMAMA_REMOTE_CLIENTAUTH=True)
    @override_settings(WARMAMA_AUTH_URL='http://www.wmmauth.com/getauth')
    @mock.patch('requests.post')
    def test_step_one_remote(self, requests_post):
        """It should send the post request with the proper arguments"""
        response = self.c.post('/clogin', {'login': 'user', 'passwd': 'pass'})
        time.sleep(0.1)  # Give the `remote_login` thread some time
        login_player = models.LoginPlayer.objects.get(login='user')

        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(response.content, {
            'ready': -1,
            'handle': login_player.pk,
            'id': 0,
        })

        # Check the `remote_login` thread made the right post request
        self.assertEqual(requests_post.call_count, 1)
        requests_post.assert_called_with(
            'http://www.wmmauth.com/getauth',
            timeout=5,
            data={
                'login': 'user',
                'passwd': 'pass',
                'handle': '%d' % login_player.pk,
                'digest': 'something',
                'url': 'http://testserver/auth',
            }
        )

    def test_step_two_not_ready(self):
        login_player = models.LoginPlayer.objects.create(
            login='user', ready=False, valid=False
        )
        response = self.c.post('/clogin', {'handle': login_player.pk})

        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(response.content, {'ready': 1, 'id': 0})

    def test_step_two_invalid(self):
        login_player = models.LoginPlayer.objects.create(
            login='user', ready=True, valid=False
        )
        response = self.c.post('/clogin', {'handle': login_player.pk})

        self.assertEqual(response.status_code, 403)
        self.assertJSONEqual(response.content, {'ready': 2, 'id': 0})
        with self.assertRaises(models.LoginPlayer.DoesNotExist):
            models.LoginPlayer.objects.get(pk=login_player.pk)

    def test_step_two_valid(self):
        login_player = models.LoginPlayer.objects.create(
            login='user', ready=True, valid=True
        )
        response = self.c.post('/clogin', {'handle': login_player.pk})

        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(response.content, {
            "ready": 2,
            "ratings": [
                {"deviation": 0.0, "rating": 5.0, "gametype": "gt_gt_a"},
                {"deviation": 0.0, "rating": 4.0, "gametype": "gt_gt_b"}
            ],
            "id": 1,
            "profile_url_rml": "",
            "profile_url": ""
        })


class ClientLogoutTest(ClientTestCase):
    def test_mark_purgable(self):
        """It should mark finished sessions as purgable"""
        models.PurgePlayer.objects.create(player=self.player, session=self.csession)
        response = self.c.post('/clogout', {'csession': self.csession.pk})
        csession = models.PlayerSession.objects.get(pk=self.csession.pk)

        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(response.content, {'status': 1})
        self.assertTrue(csession.purgable)

    def test_delete(self):
        """It should delete sessions not marked as purgable"""
        response = self.c.post('/clogout', {'csession': self.csession.pk})

        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(response.content, {'status': 1})
        with self.assertRaises(models.PlayerSession.DoesNotExist):
            models.PlayerSession.objects.get(pk=self.csession.pk)


class ClientConnectTest(ClientTestCase):
    def test_connect(self):
        response = self.c.post('/ccc', {
            'csession': self.csession.pk,
            'saddr': '%s:%d' % (self.ssession.ipv6, self.ssession.port)
        })
        csession = models.PlayerSession.objects.get(pk=self.csession.pk)

        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(response.content, {'ticket': csession.ticket_id})
        self.assertTrue(csession.ticket_id)


class ClientHeartbeatTest(ClientTestCase):
    def test_heartbeat(self):
        """It should update the `updated` timestamp"""
        # Force set `updated` to a time in the past and make sure it worked
        models.PlayerSession.objects.update(updated=timezone.now() - timedelta(days=1))
        csession = models.PlayerSession.objects.get(pk=self.csession.pk)
        self.assertTrue(timezone.now() - csession.updated > timedelta(minutes=1))

        response = self.c.post('/chb', {'csession': self.csession.pk})
        csession = models.PlayerSession.objects.get(pk=self.csession.pk)

        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(response.content, {'status': 1})
        self.assertTrue(timezone.now() - csession.updated < timedelta(minutes=1))


class ClientAuthenticateTest(ClientTestCase):
    def test_authenticate(self):
        login_player = models.LoginPlayer.objects.create(
            login='user', ready=False, valid=False
        )
        response = self.c.post('/auth', {
            'handle': login_player.pk,
            'digest': 'something',
            'valid': '1',
            'profile_url': 'url',
            'profile_url_rml': 'url_rml',
        })
        login_player = models.LoginPlayer.objects.get(pk=login_player.pk)

        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(response.content, {'status': 1})



import json
import six
from django.test import TestCase as DjTestCase


def loadJson(msg):
    """Read json from bytes or unicode message"""
    if isinstance(msg, six.binary_type):
        msg = six.text_type(msg, encoding='utf-8')
    return json.loads(msg)


class TestCase(DjTestCase):
    """Wrapper to handle bytestrings in JSON assertions"""

    def assertJSONEqual(self, raw, expected, msg=None):
        if isinstance(raw, six.binary_type):
            raw = six.text_type(raw, encoding='utf-8')
        super(TestCase, self).assertJSONEqual(raw, expected, msg)

    def assertJSONNotEqual(self, raw, expected, msg=None):
        if isinstance(raw, six.binary_type):
            raw = six.text_type(raw, encoding='utf-8')
        super(TestCase, self).assertJSONNotEqual(raw, expected, msg)



import json
import zlib
from base64 import urlsafe_b64encode
from datetime import datetime
from django import forms
from django.utils import timezone
from rest_framework import serializers
from rest_framework.exceptions import ValidationError as DrfValidationError
from tests import TestCase
from warmama import fields


class IPAddressFieldTest(TestCase):
    class SampleForm(forms.Form):
        ipaddr = fields.IPAddressField(unpack_ipv4=True)

    def test_ipv6(self):
        """Test strip ipv6 port"""
        # TODO: this is a non-standard way to write ipv6+port, but its what
        # qfusion does. It should be [ipv6]:port
        form = self.SampleForm({
            'ipaddr': '2001:0db8:85a3:0000:0000:8a2e:0370:7334:44400'
        })

        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data['ipaddr'], ('2001:db8:85a3::8a2e:370:7334', 44400))

    def test_ipv4(self):
        """Test strip ipv4 port"""
        form = self.SampleForm({
            'ipaddr': '192.102.2.1:44400'
        })

        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data['ipaddr'], ('192.102.2.1', 44400))

    def test_packed_ipv4(self):
        """It strip packed ipv4 port"""
        form = self.SampleForm({
            'ipaddr': '::ffff:192.102.2.1:44400'
        })

        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data['ipaddr'], ('192.102.2.1', 44400))

    def test_ipv6_noport(self):
        """Test ipv6 without port"""
        form = self.SampleForm({
            'ipaddr': '2001:0db8:85a3:0000:0000:8a2e:0370:7334'
        })

        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data['ipaddr'], ('2001:db8:85a3::8a2e:370:7334', None))

    def test_ipv4_noport(self):
        """Test ipv4 without port"""
        form = self.SampleForm({
            'ipaddr': '192.102.2.1'
        })

        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data['ipaddr'], ('192.102.2.1', None))

    def test_packed_ipv4_noport(self):
        """It packed ipv4 without port"""
        form = self.SampleForm({
            'ipaddr': '::ffff:192.102.2.1'
        })

        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data['ipaddr'], ('192.102.2.1', None))

    def test_invalid_input(self):
        """It should invalidate bad inputs"""
        form = self.SampleForm({'ipaddr': 3})

        self.assertFalse(form.is_valid())


class GzipJsonFieldTest(TestCase):
    class SampleForm(forms.Form):
        data = fields.GzipJsonField()

    def test_valid_field(self):
        data = {'a': 1, 'b': [True, "2", None]}
        packed = json.dumps(data).encode('utf-8')
        packed = zlib.compress(packed)
        packed = urlsafe_b64encode(packed).decode('ascii')
        form = self.SampleForm({'data': packed})

        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data['data'], data)


class TimestampFieldTest(TestCase):
    class SampleSerializer(serializers.Serializer):
        """Sample serializer to test TimestampField"""
        datetime = fields.TimestampField()

    class Timestamp(object):
        """Sample object that can be serialized with SampleSerializer"""
        def __init__(self, dt):
            self.datetime = dt

    def test_deserialize(self):
        """It should return correct values for sample times"""
        ts, dt = 0, datetime(1970, 1, 1, tzinfo=timezone.utc)
        serializer = self.SampleSerializer(data={'datetime': ts})
        self.assertTrue(serializer.is_valid())
        self.assertEqual(serializer.validated_data['datetime'], dt)

        ts, dt = 1420070400, datetime(2015, 1, 1, tzinfo=timezone.utc)
        serializer = self.SampleSerializer(data={'datetime': ts})
        self.assertTrue(serializer.is_valid())
        self.assertEqual(serializer.validated_data['datetime'], dt)

    def test_serialize(self):
        ts, obj = 0, self.Timestamp(datetime(1970, 1, 1, tzinfo=timezone.utc))
        serializer = self.SampleSerializer(instance=obj)
        self.assertEqual(serializer.data['datetime'], ts)

        ts, obj = 1420070400, self.Timestamp(datetime(2015, 1, 1, tzinfo=timezone.utc))
        serializer = self.SampleSerializer(instance=obj)
        self.assertEqual(serializer.data['datetime'], ts)

    def test_naive(self):
        obj = self.Timestamp(datetime(1970, 1, 1))
        serializer = self.SampleSerializer(instance=obj)

        with self.assertRaises(DrfValidationError):
            serializer.data



"""Warmama database models"""
import uuid
from django.db import models
from django.utils.encoding import python_2_unicode_compatible


##########
# Game Models
##########


@python_2_unicode_compatible
class Award(models.Model):
    """Award

    Attributes:
        name (str): Name of the map.
    """
    name = models.CharField(max_length=64, unique=True)

    class Meta:
        db_table = 'awards'

    def __str__(self):
        return '{0}'.format(self.name)


@python_2_unicode_compatible
class Achievement(models.Model):
    """Achievement

    Attributes:
        name (str): Name of the achievement
        description (str): Description of the achievement
        numgotten (int): (default: 0) #TODO no idea what this is for
    """
    name = models.CharField(max_length=64, unique=True)
    description = models.CharField(max_length=128)
    numgotten = models.IntegerField(default=0)

    class Meta:
        db_table = 'achievements'

    def __str__(self):
        return '{0}'.format(self.name)


@python_2_unicode_compatible
class Gametype(models.Model):
    """Gametype

    Attributes:
        name (str): Abbreviated name of the gametype.
        description (str): Long name of the gametype.

    Reverse lookup attributes:
        matches (QuerySet): MatchResult objects for every match of the gametype
    """
    name = models.CharField(max_length=16, unique=True)
    description = models.CharField(max_length=32)

    class Meta:
        db_table = 'gametypes'

    def __str__(self):
        return '{0}'.format(self.name)


@python_2_unicode_compatible
class Map(models.Model):
    """Map

    Attributes:
        mapname (str): Name of the map

    Reverse lookup attributes:
        matches (QuerySet): MatchResult objects for every match on the map
    """
    mapname = models.CharField(max_length=64, unique=True)

    class Meta:
        db_table = 'mapnames'

    def __str__(self):
        return '{0}'.format(self.mapname)


@python_2_unicode_compatible
class Weapon(models.Model):
    """Weapon

    Attributes:
        name (str): Abbreviated name of the weapon.
        fullname (str): Long name of the weapon.
    """
    name = models.CharField(max_length=2, unique=True)
    fullname = models.CharField(max_length=16)

    class Meta:
        db_table = 'weapons'

    def __str__(self):
        return '{0}'.format(self.name)


@python_2_unicode_compatible
class Player(models.Model):
    """Player class

    Attributes:
        created (datetime): Time when the record was created
        updated (datetime): Time when the record was last modified
        login (str): Login username. This field must be unique
        nickname (str): Ingame nickname with color codes
        ip (str): IPv4 address
        ipv6 (str): IPv6 address
        location (str): Two letter country code
        banned (bool): Player is banned (default: False)

    Reverse lookup attributes:
        session (PlayerSession): Session the player is connected with
        stats (QuerySet): PlayerStat objects for the player
        achievements (QuerySet): PlayerAchievement objects for the player
        matches (QuerySet): MatchPlayer objects for the player
    """
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)
    login = models.CharField(max_length=64, unique=True)
    nickname = models.CharField(max_length=64, blank=True)
    ip = models.CharField(max_length=22, blank=True)
    ipv6 = models.CharField(max_length=54, blank=True)
    location = models.CharField(max_length=2, blank=True, db_index=True)
    banned = models.BooleanField(default=False)

    class Meta:
        db_table = 'players'

    def __str__(self):
        return '{0}'.format(self.login)


@python_2_unicode_compatible
class Server(models.Model):
    """Server class

    Attributes:
        created (datetime): Time when the record was created
        updated (datetime): Time when the record was last modified
        login (str): Login username. This field must be unique
        regip (str): IPv4 address server registered with
        regipv6 (str): IPv6 address server registered with
        hostname (str): Hostname
        ip (str): IPv4 address server logged in with TODO: will this ever differ from regip?
        ipv6 (str): IPv6 address server logged in with TODO: will this ever differ from regipv6?
        location (str): Two letter country code
        banned (bool): Banned (default: False)
        demos_baseurl (str): `sv_uploads_demos_baseurl` value for server

    Reverse lookup attributes:
        session (ServerSession): Session the server is connected with
        matches (QuerySet): MatchResult objects for every match the server hosted
    """
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)
    login = models.CharField(max_length=64, unique=True)
    regip = models.CharField(max_length=22, blank=True, db_index=True)
    regipv6 = models.CharField(max_length=54, blank=True)
    hostname = models.CharField(max_length=64, blank=True)
    ip = models.CharField(max_length=22, blank=True, db_index=True)
    ipv6 = models.CharField(max_length=54, blank=True)
    location = models.CharField(max_length=2, blank=True, db_index=True)
    banned = models.BooleanField(default=False)
    demos_baseurl = models.CharField(max_length=128, blank=True)

    class Meta:
        db_table = 'servers'

    def __str__(self):
        return '{0}'.format(self.regip)


##########
# Sessions and Auth
##########


@python_2_unicode_compatible
class ServerSession(models.Model):
    """Server session class

    Attributes:
        created (datetime): Time when the record was created
        updated (datetime): Time when the record was last modified
        user (Server): Server the session belongs to
        ip (str): IPv4 address
        ipv6 (str): IPv6 address
        digest (str): #TODO Used for remote authentication?
        port (int): Port game is being served on.
        next_match_uuid (str): #TODO No idea what this is for
    """
    created = models.DateTimeField(auto_now_add=True, db_index=True)
    updated = models.DateTimeField(auto_now=True)
    user = models.OneToOneField('Server', related_name='session', db_index=True)
    ip = models.CharField(max_length=22, blank=True, db_index=True)
    ipv6 = models.CharField(max_length=54, blank=True)
    digest = models.CharField(default=uuid.uuid4, max_length=36)
    port = models.IntegerField()
    next_match_uuid = models.CharField(max_length=36, blank=True, unique=True)

    class Meta:
        db_table = 'sessions_server'

    def __str__(self):
        return '{0}'.format(self.pk)


@python_2_unicode_compatible
class PlayerSession(models.Model):
    """Player session class

    If a player is requesting a connection, `ticket_id`, `ticket_server`, and
    `ticket_expiration` will be defined and `server_session` will be `None`.

    If a player has connected to a server `ticket_id`, `ticket_server`, and
    `ticket_expiration` will be `None` and `server_session` will be defined.

    If a player has disconnected, the session is over. All the fields will be
    `None` and `purgable` will be `True`.

    Attributes:
        created (datetime): Time when the record was created
        updated (datetime): Time when the record was last modified
        user (Player): Player
        ip (str): IPv4 address
        ipv6 (str): IPv6 address
        digest (str): #TODO Used for remote authentication?
        ticket_id (int): Identifier for the ticket
        ticket_server (ServerSession): ServerSession that client requested to connect to
        ticket_expiration (datetime): Time when the ticket expires
        server_session (ServerSession): ServerSession that client is connected to
        purgable (bool): True if the session may be purged (default: False)

    Reverse lookup attributes:
        purge_player (PurgePlayer): Purge object for the session
    """
    created = models.DateTimeField(auto_now_add=True, db_index=True)
    updated = models.DateTimeField(auto_now=True)
    user = models.OneToOneField('Player', related_name='session')
    ip = models.CharField(max_length=22, blank=True)
    ipv6 = models.CharField(max_length=54, blank=True)
    digest = models.CharField(default=uuid.uuid4, max_length=32)
    ticket_id = models.IntegerField(default=0, blank=True, null=True)
    ticket_server = models.ForeignKey('ServerSession', blank=True, null=True, db_column='ticket_server', related_name='+', on_delete=models.SET_NULL, db_index=True)
    ticket_expiration = models.DateTimeField(blank=True, null=True)
    server_session = models.ForeignKey('ServerSession', blank=True, null=True, db_column='server_session', related_name='+', on_delete=models.SET_NULL, db_index=True)
    purgable = models.BooleanField(default=False)

    class Meta:
        db_table = 'sessions_player'

    def __str__(self):
        return '{0}'.format(self.pk)


@python_2_unicode_compatible
class PurgePlayer(models.Model):
    """Purgable player sessions

    Attributes:
        created (datetime): Time when the record was created
        updated (datetime): Time when the record was last modified
        session (PlayerSession): PlayerSession that may be purged
        player (Player): Player for the session
        server_session (ServerSession): ServerSession that marked as purgable
    """
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)
    session = models.OneToOneField('PlayerSession', related_name='purge_player', db_index=True)
    player = models.OneToOneField('Player', related_name='+', db_index=True)
    server_session = models.ForeignKey('ServerSession', blank=True, null=True, db_column='server_session', related_name='+', db_index=True, on_delete=models.SET_NULL)

    class Meta:
        db_table = 'purge_players'

    def __str__(self):
        return '{0}'.format(self.pk)


@python_2_unicode_compatible
class LoginPlayer(models.Model):
    """Login players

    Attributes:
        created (datetime): Time when record was created
        login (str): Player's login name
        ready (bool): Ready state (default: True)
        valid (bool): Valid state (default: True)
        profile_url (str): Url to player profile
        profile_url_rml (str): Url to player profile in rml format
    """
    created = models.DateTimeField(auto_now_add=True)
    login = models.CharField(max_length=64, unique=True)
    ready = models.BooleanField(default=False)
    valid = models.BooleanField(default=False)
    profile_url = models.CharField(max_length=255, blank=True)
    profile_url_rml = models.CharField(max_length=255, blank=True)

    class Meta:
        db_table = 'login_players'

    def __str__(self):
        return '{0}'.format(self.pk)


##########
# Player Results
##########


@python_2_unicode_compatible
class PlayerStat(models.Model):
    """Overall stats for a (player, gametype) pair

    Attributes:
        player (Player): Player stats belong to
        created (datetime): Time the record was first created
        updated (datetime): Time the record was last modified
        gametype (Gametype): Gametype the stat is for
        wins (int): Number of wins (default: 0)
        losses (int): Number of losses (default: 0)
        quits (int): Number of quits (default: 0)
        rating (Decimal): Player's rating value (mu) for the gametype (default: 0)
        deviation (Decimal): Player's rating deviation (sigma) for the gametype (default: 0)
    """
    player = models.ForeignKey('Player', related_name='stats')
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True, db_index=True)
    gametype = models.ForeignKey('Gametype', related_name='+')
    wins = models.IntegerField(default=0)
    losses = models.IntegerField(default=0)
    quits = models.IntegerField(default=0)
    rating = models.DecimalField(default=0, max_digits=8, decimal_places=2)
    deviation = models.DecimalField(default=0, max_digits=8, decimal_places=2)

    class Meta:
        db_table = 'player_stats'
        unique_together = ('player', 'gametype')

    def __str__(self):
        return '{0} {1}'.format(self.gametype_id, self.player_id)


@python_2_unicode_compatible
class PlayerAchievement(models.Model):
    """Achievements a player has won

    Attributes:
        player (Player): Player
        created (datetime): Time the record was first created
        updated (datetime): Time the record was last modified
        achievement (Achievement): Achievement
    """
    player = models.ForeignKey('Player', related_name='achievements', db_index=True)
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)
    achievement = models.ForeignKey('Achievement', related_name='+')

    class Meta:
        db_table = 'player_achievements'

    def __str__(self):
        return '{0} {1}'.format(self.achievement_id, self.player_id)


##########
# Match Results
##########


@python_2_unicode_compatible
class MatchResult(models.Model):
    """Match Result

    Attributes:
        created (datetime): Time the record was first created
        updated (datetime): Time the record was last modified
        server (Server): Server the match was hosted on
        gametype (Gametype): Gametype for the match
        uuid (str): Unique identifier string for the match result (default: uuid.uuid4)
        instagib (bool): True if match was instagib (default: False)
        teamgame (bool): True if match was teamgame (default: False)
        map_id (int): Pointer to map's pk
        timelimit (int): Time limit for the match (default: 0)
        scorelimit (int): Score limit for the match (default: 0)
        gamedir (str): `fs_game` value for the match
        matchtime (int): Length of the match in seconds (default: 0)
        utctime (datetime): Time the match was ended
        winner_team (MatchTeam): Winning MatchTeam
        winner_player (MatchPlayer): Winning MatchPlayer
        demo_filename (str): Name of demo file for the match

    Reverse lookup attributes:
        teams (QuerySet): MatchTeam objects in the match
        players (QuerySet): MatchPlayer objects in the match
        frags (QuerySet): MatchFrag objects in the match
    """
    created = models.DateTimeField(auto_now_add=True, db_index=True)
    updated = models.DateTimeField(auto_now=True)
    server = models.ForeignKey('Server', related_name='matches', db_index=True)
    gametype = models.ForeignKey('Gametype', related_name='matches', db_index=True)
    uuid = models.CharField(default=uuid.uuid4, max_length=36, unique=True)
    instagib = models.BooleanField(default=False)
    teamgame = models.BooleanField(default=False)
    map = models.ForeignKey('Map', related_name='matches')
    timelimit = models.IntegerField(default=0)
    scorelimit = models.IntegerField(default=0)
    gamedir = models.CharField(max_length=64, blank=True)
    matchtime = models.IntegerField(default=0)
    utctime = models.DateTimeField()
    winner_team = models.ForeignKey('MatchTeam', related_name='+', blank=True, null=True, db_index=True)
    winner_player = models.ForeignKey('MatchPlayer', related_name='+', blank=True, null=True, db_index=True)
    demo_filename = models.CharField(max_length=128, blank=True)

    class Meta:
        db_table = 'match_results'

    def __str__(self):
        return '{0}'.format(self.pk)


@python_2_unicode_compatible
class MatchTeam(models.Model):
    """Team in a match

    Attributes:
        matchresult (MatchResult): MatchResult the team was a member of
        name (str): Name of the team
        score (int): Score earned by the team (default: 0)

    Reverse lookup attributes:
        players (QuerySet): MatchPlayer objects in the match
    """
    matchresult = models.ForeignKey('MatchResult', related_name='teams', db_index=True)
    name = models.CharField(max_length=64)
    score = models.IntegerField(default=0)

    class Meta:
        db_table = 'match_teams'
        unique_together = ('matchresult', 'name')

    def __str__(self):
        return '{0}'.format(self.name)


@python_2_unicode_compatible
class MatchPlayer(models.Model):
    """Player results in a match

    Attributes:
        player (Player): Player
        matchresult (MatchResult): MatchResult
        matchteam (MatchTeam): MatchTeam
        name (str): Player's name in match TODO: is this login or nickname?
        score (int): Score earned (default: 0)
        frags (int): Frags made (default: 0)
        deaths (int): Times killed (default: 0)
        teamkills (int): Betrayal kills (default: 0)
        suicides (int): Suicides made (default: 0)
        numrounds (int): Rounds played in the match (default: 0)
        ga_taken (int): Green armors taken (default: 0)
        ya_taken (int): Yellow armors taken (default: 0)
        ra_taken (int): Red armors taken (default: 0)
        mh_taken (int): Mega healths taken (default: 0)
        uh_taken (int): Ultra healths taken (default: 0)
        quads_taken (int): Quad damages taken (default: 0)
        shells_taken (int): Wsw shells taken (default: 0)
        bombs_planted (int): Bombs planted (default: 0)
        bombs_defused (int): Bombs defused (default: 0)
        flags_capped (int): Flags captured (default: 0)
        matchtime (int): Playtime in seconds (default: 0)
        oldrating (Decimal): Player's rating before the match (default: 0)
        newrating (Decimal): Player's rating after the match (default: 0)
    """
    # TODO: add dmg, health, armor given/taken
    player = models.ForeignKey('Player', related_name='matches', db_index=True)
    matchresult = models.ForeignKey('MatchResult', related_name='players', db_index=True)
    matchteam = models.ForeignKey('MatchTeam', related_name='players', blank=True, null=True, db_index=True)
    name = models.CharField(max_length=64, blank=True)
    score = models.IntegerField(default=0)
    frags = models.IntegerField(default=0)
    deaths = models.IntegerField(default=0)
    teamkills = models.IntegerField(default=0)
    suicides = models.IntegerField(default=0)
    numrounds = models.IntegerField(default=0)
    ga_taken = models.IntegerField(default=0)
    ya_taken = models.IntegerField(default=0)
    ra_taken = models.IntegerField(default=0)
    mh_taken = models.IntegerField(default=0)
    uh_taken = models.IntegerField(default=0)
    quads_taken = models.IntegerField(default=0)
    shells_taken = models.IntegerField(default=0)
    bombs_planted = models.IntegerField(default=0)
    bombs_defused = models.IntegerField(default=0)
    flags_capped = models.IntegerField(default=0)
    matchtime = models.IntegerField(default=0)
    oldrating = models.DecimalField(default=0, max_digits=8, decimal_places=2)
    newrating = models.DecimalField(default=0, max_digits=8, decimal_places=2)

    class Meta:
        db_table = 'match_players'
        unique_together = ('player', 'matchresult')

    def __str__(self):
        return '{0}'.format(self.player_id)


@python_2_unicode_compatible
class MatchAward(models.Model):
    """Player awards for a match

    Attributes:
        player (Player): Player
        matchresult (MatchResult): MatchResult
        award (Award): Award
        count (int): Number of times earned in the match (default: 0)
    """
    player = models.ForeignKey('Player', related_name='+', db_index=True)
    matchresult = models.ForeignKey('MatchResult', related_name='+', db_index=True)
    award = models.ForeignKey('Award', related_name='+', db_index=True)
    count = models.IntegerField(default=0)

    class Meta:
        db_table = 'match_awards'

    def __str__(self):
        return '{0}'.format(self.pk)


@python_2_unicode_compatible
class MatchWeapon(models.Model):
    """Weapon statistics for a player match

    Attributes:
        player (Player): Player
        matchresult (MatchResult): MatchResult
        weapon (Weapon): Weapon
        shots_strong (int): Number of shots made with strong ammo (default: 0)
        hits_stong (int): Number of hits made with strong ammo (default: 0)
        dmg_strong (int): Total damage dealt with strong ammo (default: 0)
        frags_strong (int): Number of kills made with strong ammo (default: 0)
        acc_strong (Decimal): Accuracy with strong ammo, hits / shots (default: 0)
        shots_weak (int): Number of shots made with weak ammo (default: 0)
        hits_weak (int): Number of hits made with weak ammo (default: 0)
        dmg_weak (int): Total damage dealt with weak ammo (default: 0)
        frags_weak (int): Number of kills made with weak ammo (default: 0)
        acc_weak (Decimal): Accuracy with weak ammo, hits / shots (default: 0)
    """
    player = models.ForeignKey('Player', related_name='+', db_index=True)
    matchresult = models.ForeignKey('MatchResult', related_name='+', db_index=True)
    weapon = models.ForeignKey('Weapon', related_name='+', db_index=True)
    shots_strong = models.IntegerField(default=0)
    hits_strong = models.IntegerField(default=0)
    dmg_strong = models.IntegerField(default=0)
    frags_strong = models.IntegerField(default=0)
    acc_strong = models.DecimalField(default=0, max_digits=5, decimal_places=2)
    shots_weak = models.IntegerField(default=0)
    hits_weak = models.IntegerField(default=0)
    dmg_weak = models.IntegerField(default=0)
    frags_weak = models.IntegerField(default=0)
    acc_weak = models.DecimalField(default=0, max_digits=5, decimal_places=2)

    class Meta:
        db_table = 'match_weapons'

    def __str__(self):
        return '{0}'.format(self.pk)


@python_2_unicode_compatible
class MatchFrag(models.Model):
    """Frags made in a match

    Attributes:
        created (datetime): Time the record was created
        matchresult (MatchResult): MatchResult
        attacker (MatchPlayer): attacker
        victim (MatchPlayer): victim
        weapon (Weapon): Weapon used in the frag, may be None
        time (int): match time in seconds the frag was made (default: 0)
    """
    created = models.DateTimeField(auto_now_add=True)
    matchresult = models.ForeignKey('MatchResult', related_name='frags', db_index=True)
    attacker = models.ForeignKey('Player', related_name='+', db_index=True)
    victim = models.ForeignKey('Player', related_name='+', db_index=True)
    weapon = models.ForeignKey('Weapon', blank=True, null=True, db_index=True)
    time = models.IntegerField(default=0)

    class Meta:
        db_table = 'frag_log'

    def __str__(self):
        return '{0}'.format(self.pk)


##########
# Race Results
##########


@python_2_unicode_compatible
class RaceRun(models.Model):
    """Race run

    Attributes:
        created (datetime): Time when the record was created
        map (Map): Map
        server (Server): Server
        player (Player): Player
        utctime (datetime): Time when the race was finished

    Reverse lookup attributes:
        sectors (QuerySet): RaceSectors for the run
    """
    created = models.DateTimeField(auto_now_add=True, db_index=True)
    map = models.ForeignKey('Map', related_name='+', db_index=True)
    server = models.ForeignKey('Server', related_name='+', db_index=True)
    player = models.ForeignKey('Player', related_name='+', db_index=True)
    utctime = models.DateTimeField()

    class Meta:
        db_table = 'race_runs'

    def __str__(self):
        return '{0}'.format(self.pk)


@python_2_unicode_compatible
class RaceSector(models.Model):
    """Race sector / checkpoint

    Attributes:
        created (datetime): Time the record was created
        run (RaceRun): RaceRun this belongs to
        sector (int): Id of the sector on the map, `-1` indicates the total
            race time.
        time (int): Time when the sector was crossed in milliseconds (default: 0)
    """
    created = models.DateTimeField(auto_now_add=True)
    run = models.ForeignKey('RaceRun', related_name='sectors', db_index=True)
    sector = models.IntegerField()
    time = models.IntegerField(default=0, db_index=True)

    class Meta:
        db_table = 'race_sectors'
        unique_together = ('run', 'sector')

    def __str__(self):
        return '{0}'.format(self.pk)



"""Module for warmama exceptions"""


class WarmamaException(Exception):
    """Base class for Warmama exceptions

    Attributes:
        status (int): Response status a view should return if this exception
            is encountered.
    """
    status = 500


class BadRequest(WarmamaException):
    """Exception raised for invalid form data or otherwise 400 responses"""
    status = 400


class Forbidden(WarmamaException):
    """Exception raised on authentication or permission failure"""
    status = 403



from datetime import datetime
from django.utils import timezone
from tests import TestCase
from warmama import models, serializers


class Data(object):
    """Class for creating serializable objects, it maps kwargs to attributes"""
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)


class MatchTeamSerializerTest(TestCase):
    def test_serialize(self):
        data = {'name': 'team', 'score': 10, 'index': 1}
        serializer = serializers.MatchTeamSerializer(Data(**data))

        self.assertEqual(serializer.data, data)

    def test_deserialize(self):
        data = {'name': 'team', 'score': 10, 'index': 1}
        serializer = serializers.MatchTeamSerializer(data=data)
        expected = {'name': 'team', 'score': 10, 'index': 1}

        serializer.is_valid()
        self.assertFalse(serializer.errors)
        self.assertEqual(serializer.validated_data, expected)


class MatchWeaponSerializerTest(TestCase):
    def test_serialize(self):
        data = Data(
            weapon='pg',
            acc_strong=2.3,
            dmg_strong=1,
            acc_weak=5.6,
            dmg_weak=7,
        )
        serializer = serializers.MatchWeaponSerializer(data)
        expected = {
            'weapon': 'pg',
            'acc_strong': 2.3,
            'dmg_strong': 1,
            'acc_weak': 5.6,
            'dmg_weak': 7,
        }

        self.assertEqual(serializer.data, expected)

    def test_deserialize(self):
        data = {
            'weapon': 'pg',
            'acc_strong': 2.3,
            'dmg_strong': 1,
            'acc_weak': 5.6,
            'dmg_weak': 7,
        }
        serializer = serializers.MatchWeaponSerializer(data=data)
        expected = data

        serializer.is_valid()
        self.assertFalse(serializer.errors)
        self.assertEqual(serializer.validated_data, expected)


class MatchWeaponSetSerializerTest(TestCase):
    def setUp(self):
        self.weapon_data = {
            'acc_strong': 2.3,
            'dmg_strong': 1,
            'acc_weak': 5.6,
            'dmg_weak': 7,
        }

    def test_serialize(self):
        weapons = [Data(**self.weapon_data.copy()) for _ in range(3)]
        weapons[0].weapon = 'pg'
        weapons[1].weapon = 'rl'
        weapons[2].weapon = 'eb'
        serializer = serializers.MatchWeaponSetSerializer(instance=weapons)
        expected = {name: self.weapon_data for name in ('pg', 'rl', 'eb')}

        self.assertEqual(serializer.data, expected)

    def test_deserialize(self):
        data = {name: self.weapon_data.copy() for name in ('pg', 'rl', 'eb')}
        serializer = serializers.MatchWeaponSetSerializer(data=data)

        serializer.is_valid()
        self.assertFalse(serializer.errors)
        for weapon in serializer.validated_data:
            self.assertIn(weapon.pop('weapon'), ('pg', 'rl', 'eb'))
            self.assertEqual(weapon, self.weapon_data)

    def test_invalid_weapon(self):
        data = {'long': self.weapon_data.copy()}
        serializer = serializers.MatchWeaponSetSerializer(data=data)

        self.assertFalse(serializer.is_valid())

    def test_invalid_weaponset(self):
        serializer = serializers.MatchWeaponSetSerializer(data=3)

        self.assertFalse(serializer.is_valid())


class MatchAwardSerializerTest(TestCase):
    def test_serialize(self):
        data = Data(name='award', count=2)
        serializer = serializers.MatchAwardSerializer(data)
        expected = {'name': 'award', 'count': 2}

        self.assertEqual(serializer.data, expected)

    def test_deserialize(self):
        data = {'name': 'award', 'count': 2}
        serializer = serializers.MatchAwardSerializer(data=data)
        expected = {'name': 'award', 'count': 2}

        serializer.is_valid()
        self.assertFalse(serializer.errors)
        self.assertEqual(serializer.validated_data, expected)


class MatchFragSerializerTest(TestCase):
    def setUp(self):
        self.victim = models.Player.objects.create(
            login='victim',
            ip='132.233.139.177',
        )
        self.vsession = models.PlayerSession.objects.create(
            user_id=self.victim.pk,
            ip=self.victim.ip,
        )

    def test_serialize(self):
        data = Data(victim=self.vsession, weapon='pg')
        serializer = serializers.MatchFragSerializer(data)
        expected = {
            'victim': self.vsession.pk,
            'weapon': 'pg',
        }

        self.assertEqual(serializer.data, expected)

    def test_deserialize(self):
        data = {
            'victim': self.vsession.pk,
            'weapon': 'pg',
        }
        serializer = serializers.MatchFragSerializer(data=data)
        expected = {
            'victim': self.vsession,
            'weapon': 'pg',
        }

        serializer.is_valid()
        self.assertFalse(serializer.errors)
        self.assertEqual(serializer.validated_data, expected)


class MatchPlayerSerializerTest(TestCase):
    def setUp(self):
        self.player = models.Player.objects.create(
            login='player',
            ip='132.233.139.177',
        )
        self.csession = models.PlayerSession.objects.create(
            user_id=self.player.pk,
            ip=self.player.ip,
        )

    def test_serialize(self):
        data = {
            'final': 1,
            'sessionid': self.csession,
            'awards': [
                Data(name='a', count=2),
                Data(name='b', count=4),
            ],
        }
        serializer = serializers.MatchPlayerSerializer(Data(**data))
        expected = {
            'final': 1,
            'sessionid': self.csession.pk,
            'awards': [
                {'name': 'a', 'count': 2},
                {'name': 'b', 'count': 4},
            ],
        }

        self.assertEqual(serializer.data, expected)

    def test_deserialize(self):
        data = {
            'final': 1,
            'sessionid': self.csession.pk,
            'awards': [
                {'name': 'a', 'count': 2},
                {'name': 'b', 'count': 4},
            ],
        }
        serializer = serializers.MatchPlayerSerializer(data=data)
        expected = {
            'final': 1,
            'sessionid': self.csession,
            'awards': [
                {'name': 'a', 'count': 2},
                {'name': 'b', 'count': 4},
            ],
        }

        serializer.is_valid()
        self.assertFalse(serializer.errors)
        self.assertEqual(serializer.validated_data, expected)


class MatchRaceRunSerializerTest(TestCase):
    def setUp(self):
        self.player = models.Player.objects.create(
            login='player',
            ip='132.233.139.177',
        )
        self.csession = models.PlayerSession.objects.create(
            user_id=self.player.pk,
            ip=self.player.ip,
        )

    def test_serialize(self):
        data = Data(
            session_id=self.csession,
            utctime=datetime(1970, 1, 1, tzinfo=timezone.utc),
            times=[1, 2, 3],
        )
        serializer = serializers.MatchRaceRunSerializer(data)
        expected = {
            'session_id': self.csession.pk,
            'timestamp': 0,
            'times': [1, 2, 3],
        }

        self.assertEqual(serializer.data, expected)

    def test_deserialize(self):
        data = {
            'session_id': self.csession.pk,
            'timestamp': 0,
            'times': [1, 2, 3],
        }
        serializer = serializers.MatchRaceRunSerializer(data=data)
        expected = {
            'session_id': self.csession,
            'utctime': datetime(1970, 1, 1, tzinfo=timezone.utc),
            'times': [1, 2, 3],
        }

        serializer.is_valid()
        self.assertFalse(serializer.errors)
        self.assertEqual(serializer.validated_data, expected)


class MatchResultSerializerTest(TestCase):
    def setUp(self):
        self.player = models.Player.objects.create(
            login='player',
            ip='132.233.139.177',
        )
        self.csession = models.PlayerSession.objects.create(
            user_id=self.player.pk,
            ip=self.player.ip,
        )

    def test_deserialize(self):
        data = {
            'gametype': 'ca',
            'teamgame': True,
            'map': 'wca1',
            'hostname': 'host',
            'timeplayed': 1000,
            'timelimit': 999,
            'scorelimit': 998,
            'instagib': True,
            'teamgame': True,
            'racegame': False,
            'timestamp': 0,
            'gamedir': 'basewsw',
            'demo_filename': 'demo',
        }
        serializer = serializers.MatchResultSerializer(data=data)
        expected = {
            'gametype': 'ca',
            'teamgame': True,
            'map': 'wca1',
            'hostname': 'host',
            'matchtime': 1000,
            'timelimit': 999,
            'scorelimit': 998,
            'instagib': True,
            'teamgame': True,
            'racegame': False,
            'utctime': datetime(1970, 1, 1, tzinfo=timezone.utc),
            'gamedir': 'basewsw',
            'demo_filename': 'demo',
        }

        serializer.is_valid()
        self.assertFalse(serializer.errors)
        self.assertEqual(serializer.validated_data, expected)


class MatchSerializerTest(TestCase):
    def setUp(self):
        self.attacker = models.Player.objects.create(
            login='attacker',
            ip='132.233.139.177',
        )
        self.asession = models.PlayerSession.objects.create(
            user_id=self.attacker.pk,
            ip=self.attacker.ip,
        )
        self.victim = models.Player.objects.create(
            login='victim',
            ip='132.139.233.177',
        )
        self.vsession = models.PlayerSession.objects.create(
            user_id=self.victim.pk,
            ip=self.victim.ip,
        )
        self.server = models.Server.objects.create(
            login='server',
            ip='1:2:3:4:5:6:7:8',
        )
        self.ssession = models.ServerSession.objects.create(
            user_id=self.server.pk,
            ip=self.server.ip,
            port=44400,
        )

        self.data = {
            'match': {
                'gametype': 'ca',
                'teamgame': True,
                'racegame': False,
                'map': 'wca1',
                'hostname': 'host',
                'timeplayed': 100,
                'timestamp': 0,
                'gamedir': 'basewsw',
            },
            'players': [
                {
                    'final': 1,
                    'team': 1,
                    'sessionid': self.asession.pk,
                    'frags': 2,
                    'deaths': 0,
                    'awards': [
                        {'name': 'a', 'count': 2},
                        {'name': 'b', 'count': 3},
                    ],
                    'weapons': {
                        'pg': {'dmg_strong': 10},
                        'rl': {'dmg_strong': 20},
                    },
                    'log_frags': [
                        {
                            'victim': self.vsession.pk,
                            'weapon': 'pg',
                            'time': 1,
                        }, {
                            'victim': self.vsession.pk,
                            'weapon': 'rl',
                            'time': 2,
                        }
                    ],
                }, {
                    'final': 1,
                    'team': 2,
                    'frags': 0,
                    'deaths': 2,
                    'sessionid': self.vsession.pk,
                    'awards': [{'name': 'b', 'count': 4}],
                }
            ],
            'teams': [
                {'name': 'Alpha', 'score': 2, 'index': 1},
                {'name': 'Beta', 'score': 0, 'index': 2},
            ],
        }
        self.serializer = serializers.MatchSerializer(data=self.data)
        self.serializer.is_valid()
        self.serializer.save(server_session=self.ssession)

    def test_valid(self):
        self.assertFalse(self.serializer.errors)

    def test_create_base_models(self):
        """It should create base models if they don't exist"""
        models.Gametype.objects.get(name='ca')
        models.Map.objects.get(mapname='wca1')
        models.Weapon.objects.get(name='pg')
        models.Weapon.objects.get(name='rl')

    def test_matchresult(self):
        """Ensure the matchresult is properly created"""
        matchresult = models.MatchResult.objects.get(server=self.server)
        self.assertEqual(matchresult.gametype.name, 'ca')
        self.assertEqual(matchresult.map.mapname, 'wca1')
        self.assertEqual(matchresult.gamedir, 'basewsw')

    def test_matchplayers(self):
        """Ensure players are properly created"""
        attacker = models.MatchPlayer.objects.get(player=self.attacker)
        victim = models.MatchPlayer.objects.get(player=self.victim)
        self.assertEqual(attacker.matchteam.name, 'Alpha')
        self.assertEqual(attacker.deaths, 0)
        self.assertEqual(attacker.frags, 2)
        self.assertEqual(victim.matchteam.name, 'Beta')
        self.assertEqual(victim.deaths, 2)
        self.assertEqual(victim.frags, 0)

    def test_matchawards(self):
        """Ensure awards are properly created"""
        self.assertEqual(models.MatchAward.objects.filter(player=self.attacker).count(), 2)
        self.assertEqual(models.MatchAward.objects.filter(player=self.victim).count(), 1)

    def test_matchfrags(self):
        """Ensure frags are properly created"""
        models.MatchFrag.objects.get(attacker=self.attacker, weapon__name='pg')
        models.MatchFrag.objects.get(attacker=self.attacker, weapon__name='rl')

    def test_matchweapons(self):
        """Ensure weapons are properly created"""
        models.MatchWeapon.objects.get(player=self.attacker, weapon__name='pg', dmg_strong=10)
        models.MatchWeapon.objects.get(player=self.attacker, weapon__name='rl', dmg_strong=20)


class RaceMatchSerializerTest(TestCase):
    def setUp(self):
        self.player = models.Player.objects.create(
            login='player',
            ip='132.233.139.177',
        )
        self.csession = models.PlayerSession.objects.create(
            user_id=self.player.pk,
            ip=self.player.ip,
        )
        self.server = models.Server.objects.create(
            login='server',
            ip='1:2:3:4:5:6:7:8',
        )
        self.ssession = models.ServerSession.objects.create(
            user_id=self.server.pk,
            ip=self.server.ip,
            port=44400,
        )

        self.data = {
            'match': {
                'gametype': 'race',
                'teamgame': False,
                'racegame': True,
                'map': 'wrace1',
                'hostname': 'host',
                'timeplayed': 100,
                'timestamp': 0,
                'gamedir': 'basewsw',
            },
            'runs': [{
                'session_id': self.csession.pk,
                'timestamp': 0,
                'times': [1, 2, 3],
            }, {
                'session_id': self.csession.pk,
                'timestamp': 0,
                'times': [2, 3, 4],
            }]
        }
        self.serializer = serializers.MatchSerializer(data=self.data)
        self.serializer.is_valid()
        self.serializer.save(server_session=self.ssession)

    def test_runs(self):
        """It should create the proper sectors and race runs"""
        # Check first race sectors
        models.RaceSector.objects.get(sector=-1, time=3)
        models.RaceSector.objects.get(sector=0, time=1)
        models.RaceSector.objects.get(sector=1, time=2)

        # Check second race sectors
        models.RaceSector.objects.get(sector=-1, time=4)
        models.RaceSector.objects.get(sector=0, time=2)
        models.RaceSector.objects.get(sector=1, time=3)

        self.assertEqual(models.RaceRun.objects.count(), 2)
        self.assertEqual(models.RaceSector.objects.count(), 6)



from django.conf import settings as user_settings
__all__ = [
    'settings',
]


DEFAULTS = {
    # PlayerSession ticket lifetime in seconds
    'WARMAMA_TICKET_EXPIRATION': 60,

    # If `True`, ClientLogin requests have username/password validated
    # by a remote url
    'WARMAMA_REMOTE_CLIENTAUTH': False,
    'WARMAMA_AUTH_URL': 'http://remote.auth.server/getauth',

    # Server port to assume if none given
    'WARMAMA_DEFAULT_SERVER_PORT': 44400,
}


class Settings(object):
    """Settings object to map settings to properties"""
    def __getattr__(self, attr):
        if attr not in DEFAULTS:
            raise AttributeError('Invalid setting %s' % attr)

        try:
            return getattr(user_settings, attr)
        except AttributeError:
            return DEFAULTS[attr]


settings = Settings()



from __future__ import division

from django.db import transaction
from rest_framework import serializers
from warmama import fields, models


class MatchTeamSerializer(serializers.ModelSerializer):
    """MatchTeam sub-serializer for MatchResults

    Excluded fields:
        matchresult

    Extra fields:
        index (int): Team index in the match, used to associate players with teams
    """
    index = serializers.IntegerField()

    class Meta:
        model = models.MatchTeam
        exclude = ('matchresult',)


class MatchWeaponSerializer(serializers.ModelSerializer):
    """MatchWeapon sub-serializer for MatchResults

    Override fields:
        weapon (str): Name of the weapon
        acc_strong (float): Handles as a float instead of a Decimal
        acc_weak (float): Handles as a float instead of a Decimal

    Excluded fields:
        player, match_result
    """
    weapon = serializers.CharField(max_length=2)
    acc_strong = serializers.FloatField(required=False)
    acc_weak = serializers.FloatField(required=False)

    class Meta:
        model = models.MatchWeapon
        exclude = ('player', 'matchresult')


class MatchWeaponSetSerializer(serializers.BaseSerializer):
    """Wrapper for serializing match weapons in the form

    ```
    {
        "name": { stats... },
        "name": { stats... },
        ...
    }
    ```
    """
    default_error_messages = {
        'invalid': 'data must be a dictionary',
        'invalid_weapon': 'Invalid weapon data: {weapon}',
    }

    def __init__(self, *args, **kwargs):
        self.skip_invalid = kwargs.pop('skip_invalid', False)
        super(MatchWeaponSetSerializer, self).__init__(*args, **kwargs)

    def to_representation(self, obj):
        """Serialize an iterable of weapons"""
        result = {}
        for weapon in obj:
            weapon_serializer = MatchWeaponSerializer(weapon)
            data = weapon_serializer.data
            name = data.pop('weapon')
            result[name] = data

        return result

    def to_internal_value(self, data):
        """Deserialize data"""
        result = []
        try:
            iterweapons = data.items()
        except AttributeError:
            self.fail('invalid')

        for name, weapon in iterweapons:
            data = weapon.copy()
            data['weapon'] = name
            weapon_serializer = MatchWeaponSerializer(data=data)
            if weapon_serializer.is_valid():
                result.append(weapon_serializer.validated_data)
            elif not self.skip_invalid:
                self.fail('invalid_weapon', weapon=data)

        return result


class MatchAwardSerializer(serializers.ModelSerializer):
    """MatchAward sub-serializer for MatchPlayer

    Included fields:
        count

    Extra fields:
        name (str): Name of the related award
    """
    name = serializers.CharField(max_length=64)

    class Meta:
        model = models.MatchAward
        fields = ('name', 'count')


class MatchFragSerializer(serializers.ModelSerializer):
    """MatchFrag sub-serializer for MatchPlayer

    Override fields:
        victim (int): Match players based on PlayerSessions
        weapon (str): Name of the weapon

    Excluded fields:
        created, matchresult, attacker
    """
    victim = serializers.PrimaryKeyRelatedField(queryset=models.PlayerSession.objects)
    weapon = serializers.CharField(max_length=2)

    class Meta:
        model = models.MatchFrag
        exclude = ('created', 'matchresult', 'attacker')


class MatchPlayerSerializer(serializers.ModelSerializer):
    """MatchReport sub-serializer for MatchPlayer objects

    Players are looked up by PlayerSession, not Player pk

    Excluded fields:
        player, matchresult, matchteam, oldrating, newrating,
        teamkills, matchtime

    Extra fields:
        final (int): If falsey, the player quit before the match ended
        sessionid (int): pk of player session, used to identify player
        team (int): index of team this player was a member of
        teamfrags (int): alias for player.teamkills
        timeplayed (int): alias for player.matchtime
        weapons (dict): Set of weapon stats for the player, see
            MatchWeaponSetSerializer for details
        awards (iterable): Iterable of Awards for the player
        log_frags (iterable): Iterable of frags for the player
    """
    final = serializers.IntegerField()
    sessionid = serializers.PrimaryKeyRelatedField(queryset=models.PlayerSession.objects)
    team = serializers.IntegerField(source='matchteam', required=False)
    teamfrags = serializers.IntegerField(source='teamkills', required=False)
    timeplayed = serializers.IntegerField(source='matchtime', min_value=1, required=False)
    weapons = MatchWeaponSetSerializer(required=False)
    awards = MatchAwardSerializer(many=True, required=False)
    log_frags = MatchFragSerializer(many=True, required=False)

    class Meta:
        model = models.MatchPlayer
        exclude = (
            'player', 'matchresult', 'matchteam', 'oldrating',
            'newrating', 'teamkills', 'matchtime'
        )


class MatchRaceRunSerializer(serializers.Serializer):
    """MatchRaceRun sub-serializer for MatchResult

    Fields:
        session_id (int): PlayerSession for the racer
        timestamp (int): POSIX Timestamp for the time race was finished
        times (iterable): Iterable of sector times
    """
    session_id = serializers.PrimaryKeyRelatedField(queryset=models.PlayerSession.objects)
    timestamp = fields.TimestampField(source='utctime')
    times = serializers.ListField(child=serializers.IntegerField(min_value=0))


class MatchResultSerializer(serializers.ModelSerializer):
    """Match Result

    Override Fields:
        gametype (str): Maps to gametype objects by name
        map (str): Maps to Map objects by name

    Excluded Fields:
        created, updated, server, uuid, matchtime, utctime, winner_team,
        winner_player

    Extra fields:
        hostname (str): TODO I have no idea
        racegame (bool): True if the game was race
        timestamp (int): POSIX Timestamp when match finished
        timeplayed (int): Length of the match in milliseconds
    """
    gametype = serializers.CharField(max_length=16)
    map = serializers.CharField(max_length=64)
    hostname = serializers.CharField(max_length=128)
    racegame = serializers.BooleanField()
    timeplayed = serializers.IntegerField(min_value=66, source='matchtime')
    timestamp = fields.TimestampField(source='utctime')

    class Meta:
        model = models.MatchResult
        exclude = (
            'created', 'updated', 'server', 'uuid', 'matchtime', 'utctime',
            'winner_team', 'winner_player'
        )

    def create(self, validated_data):
        return super(MatchResultSerializer, self).create(validated_data)


class MatchSerializer(serializers.Serializer):
    """Match Serializer

    Fields:
        match (dict): MatchResult
        teams (iterable): MatchTeams
        players (iterable): MatchPlayers (optional)
        runs (iterable): RaceRuns (optional)
    """
    match = MatchResultSerializer()
    teams = MatchTeamSerializer(many=True, required=False)
    players = MatchPlayerSerializer(many=True, required=False)
    runs = MatchRaceRunSerializer(many=True, required=False)

    def _init_cache(self):
        """Create caches for base models"""
        self._award_cache = {}  # name: models.Award
        self._weapon_cache = {}  # name: models.Weapon

    def _create_match_result(self, matchdata, server_id):
        """Create the MatchResult model from matchdata

        This may create a Gametype and Map model if nessecary
        """
        matchdata['gametype'], _ = models.Gametype.objects.get_or_create(
            name=matchdata['gametype'],
            defaults={'description': matchdata['gametype']}
        )
        matchdata['map'], _ = models.Map.objects.get_or_create(mapname=matchdata['map'])
        return models.MatchResult.objects.create(server_id=server_id, **matchdata)

    def _create_match_teams(self, teamdata, matchresult):
        """Create Team models for the match

        Creates Team models and returns them in a dict of { teamindex : Team }
        """
        # The team pk is needed for players later, so `.bulk_create()` cannot be
        # used here
        teams = {}
        for team in teamdata:
            index = team.pop('index')
            teams[index] = models.MatchTeam.objects.create(matchresult=matchresult, **team)
        return teams

    def _create_match_players(self, playerdata, matchteams):
        """Create MatchPlayers for the match

        Returns a map { session_id : MatchPlayer }
        """
        players = {}
        for matchplayer in playerdata:
            matchplayer.pop('final')  # TODO: used in skill calculation
            player_session = matchplayer.pop('sessionid')
            matchplayer['player_id'] = player_session.user_id

            teamindex = matchplayer.pop('matchteam', None)
            if teamindex is not None:
                matchplayer['matchteam'] = matchteams[teamindex]

            players[player_session.pk] = models.MatchPlayer.objects.create(**matchplayer)

        return players

    def _create_match_awards(self, awarddata):
        """Create MatchAwards for the match

        This will create missing Award models if necessary
        """
        for matchaward in awarddata:
            name = matchaward.pop('name')
            if name not in self._award_cache:
                self._award_cache[name], _ = models.Award.objects.get_or_create(name=name)

            matchaward['award'] = self._award_cache[name]

        models.MatchAward.objects.bulk_create(
            models.MatchAward(**matchaward) for matchaward in awarddata
        )

    def _create_match_frags(self, fragdata):
        """Create MatchFrags for the match

        This will create missing Weapon models if necessary
        """
        for matchfrag in fragdata:
            name = matchfrag.pop('weapon')
            if name not in self._weapon_cache:
                self._weapon_cache[name], _ = models.Weapon.objects.get_or_create(name=name, defaults={'fullname': name})

            victim_session = matchfrag.pop('victim')
            matchfrag['victim_id'] = victim_session.user_id
            matchfrag['weapon'] = self._weapon_cache[name]

        models.MatchFrag.objects.bulk_create(
            models.MatchFrag(**matchfrag) for matchfrag in fragdata
        )

    def _create_match_weapons(self, weapondata):
        """Create MatchWeapons for the match

        This will create missing Weapon models if necessary
        """
        for matchweapon in weapondata:
            name = matchweapon.pop('weapon')
            if name not in self._weapon_cache:
                self._weapon_cache[name], _ = models.Weapon.objects.get_or_create(name=name, defaults={'fullname': name})

            matchweapon['weapon'] = self._weapon_cache[name]

        models.MatchWeapon.objects.bulk_create(
            models.MatchWeapon(**matchweapon) for matchweapon in weapondata
        )

    def _create_runs(self, rundata, map_id, server_id):
        """Create Raceruns and RaceSectors for the match"""
        sectors = []
        for racerun in rundata:
            player_session = racerun.pop('session_id')
            times = sorted(racerun.pop('times'))
            racerun['map_id'] = map_id
            racerun['server_id'] = server_id
            racerun['player_id'] = player_session.user_id
            runobj = models.RaceRun.objects.create(**racerun)

            if not times:
                continue

            # The last entry is the race time, not a checkpoint
            # Gets saved with special value `sector -1`
            sectors.append(
                models.RaceSector(run=runobj, sector=-1, time=times[-1])
            )

            # The others are indexed by their position
            sectors.extend(
                models.RaceSector(run=runobj, sector=index, time=time)
                for index, time in enumerate(times[:-1])
            )

        models.RaceSector.objects.bulk_create(sectors)

    def _get_player_data(self, playerdata, matchresult):
        """Pull the awards, frags, and weapons out of players

        Returns a tuple `(awards, frags, weapons)` where each entry is a list of
        data for that model. The returned data is ready to be used in the
        relevant `_create_*` method. The argument `playerdata` will be modified
        to have the nested values removed.
        """
        awards = []
        frags = []
        weapons = []
        for player in playerdata:
            pid = player['sessionid'].user_id
            player['matchresult'] = matchresult
            player_awards = player.pop('awards', [])
            player_frags = player.pop('log_frags', [])
            player_weapons = player.pop('weapons', [])

            # attach player / match information to the entries and append them
            awards.extend(
                dict(player_id=pid, matchresult=matchresult, **awd)
                for awd in player_awards
            )
            frags.extend(
                dict(attacker_id=pid, matchresult=matchresult, **frag)
                for frag in player_frags
            )
            weapons.extend(
                dict(player_id=pid, matchresult=matchresult, **weapon)
                for weapon in player_weapons
            )
        return awards, frags, weapons

    @transaction.atomic
    def create(self, data):
        """Save the validated_data to all the associated model objects"""
        ssession = data.get('server_session', None)
        assert isinstance(ssession, models.ServerSession), (
            'A valid server session was not provided. Serializer must be saved'
            'using `.save(server_session=ServerSessionObject)`'
        )

        assert 'players' in data or 'runs' in data, (
            'No players section or runs section in match'
        )

        assert 'teams' in data or not data['match'].get('teamgame', True), (
            'No teams section in match'
        )

        self._init_cache()
        sid = ssession.user_id
        data['match'].pop('hostname', '')  # TODO what is this used for?
        data['match'].pop('racegame', False)  # not used atm

        matchresult = self._create_match_result(data['match'], sid)
        matchteams = self._create_match_teams(data.get('teams', []), matchresult)

        players = data.get('players', [])
        awards, frags, weapons = self._get_player_data(players, matchresult)

        matchplayers = self._create_match_players(players, matchteams)
        self._create_match_awards(awards)
        self._create_match_frags(frags)
        self._create_match_weapons(weapons)

        self._create_runs(data.get('runs', []), matchresult.map_id, sid)

        # TODO process stats, ratings and set winner_team / winner_player

        # Release the cache
        self._init_cache()
        return matchresult, matchplayers



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Achievement',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, verbose_name='ID', serialize=False)),
                ('name', models.CharField(unique=True, max_length=64)),
                ('description', models.CharField(max_length=128)),
                ('numgotten', models.IntegerField(default=0)),
            ],
            options={
                'db_table': 'achievements',
            },
        ),
        migrations.CreateModel(
            name='Award',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, verbose_name='ID', serialize=False)),
                ('name', models.CharField(unique=True, max_length=64)),
            ],
            options={
                'db_table': 'awards',
            },
        ),
        migrations.CreateModel(
            name='Gametype',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, verbose_name='ID', serialize=False)),
                ('name', models.CharField(unique=True, max_length=16)),
                ('description', models.CharField(max_length=32)),
            ],
            options={
                'db_table': 'gametypes',
            },
        ),
        migrations.CreateModel(
            name='LoginPlayer',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, verbose_name='ID', serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('login', models.CharField(unique=True, max_length=64)),
                ('ready', models.BooleanField(default=False)),
                ('valid', models.BooleanField(default=False)),
                ('profile_url', models.CharField(blank=True, max_length=255)),
                ('profile_url_rml', models.CharField(blank=True, max_length=255)),
            ],
            options={
                'db_table': 'login_players',
            },
        ),
        migrations.CreateModel(
            name='Map',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, verbose_name='ID', serialize=False)),
                ('mapname', models.CharField(unique=True, max_length=64)),
            ],
            options={
                'db_table': 'mapnames',
            },
        ),
        migrations.CreateModel(
            name='MatchAward',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, verbose_name='ID', serialize=False)),
                ('count', models.IntegerField(default=0)),
                ('award', models.ForeignKey(related_name='+', to='warmama.Award')),
            ],
            options={
                'db_table': 'match_awards',
            },
        ),
        migrations.CreateModel(
            name='MatchFrag',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, verbose_name='ID', serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('time', models.IntegerField(default=0)),
            ],
            options={
                'db_table': 'frag_log',
            },
        ),
        migrations.CreateModel(
            name='MatchPlayer',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, verbose_name='ID', serialize=False)),
                ('name', models.CharField(blank=True, max_length=64)),
                ('score', models.IntegerField(default=0)),
                ('frags', models.IntegerField(default=0)),
                ('deaths', models.IntegerField(default=0)),
                ('teamkills', models.IntegerField(default=0)),
                ('suicides', models.IntegerField(default=0)),
                ('numrounds', models.IntegerField(default=0)),
                ('ga_taken', models.IntegerField(default=0)),
                ('ya_taken', models.IntegerField(default=0)),
                ('ra_taken', models.IntegerField(default=0)),
                ('mh_taken', models.IntegerField(default=0)),
                ('uh_taken', models.IntegerField(default=0)),
                ('quads_taken', models.IntegerField(default=0)),
                ('shells_taken', models.IntegerField(default=0)),
                ('bombs_planted', models.IntegerField(default=0)),
                ('bombs_defused', models.IntegerField(default=0)),
                ('flags_capped', models.IntegerField(default=0)),
                ('matchtime', models.IntegerField(default=0)),
                ('oldrating', models.DecimalField(max_digits=8, default=0, decimal_places=2)),
                ('newrating', models.DecimalField(max_digits=8, default=0, decimal_places=2)),
            ],
            options={
                'db_table': 'match_players',
            },
        ),
        migrations.CreateModel(
            name='MatchResult',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, verbose_name='ID', serialize=False)),
                ('created', models.DateTimeField(db_index=True, auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('uuid', models.CharField(unique=True, default=uuid.uuid4, max_length=36)),
                ('instagib', models.BooleanField(default=False)),
                ('teamgame', models.BooleanField(default=False)),
                ('timelimit', models.IntegerField(default=0)),
                ('scorelimit', models.IntegerField(default=0)),
                ('gamedir', models.CharField(blank=True, max_length=64)),
                ('matchtime', models.IntegerField(default=0)),
                ('utctime', models.DateTimeField()),
                ('demo_filename', models.CharField(blank=True, max_length=128)),
                ('gametype', models.ForeignKey(related_name='matches', to='warmama.Gametype')),
                ('map', models.ForeignKey(related_name='matches', to='warmama.Map')),
            ],
            options={
                'db_table': 'match_results',
            },
        ),
        migrations.CreateModel(
            name='MatchTeam',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, verbose_name='ID', serialize=False)),
                ('name', models.CharField(max_length=64)),
                ('score', models.IntegerField(default=0)),
                ('matchresult', models.ForeignKey(related_name='teams', to='warmama.MatchResult')),
            ],
            options={
                'db_table': 'match_teams',
            },
        ),
        migrations.CreateModel(
            name='MatchWeapon',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, verbose_name='ID', serialize=False)),
                ('shots_strong', models.IntegerField(default=0)),
                ('hits_strong', models.IntegerField(default=0)),
                ('dmg_strong', models.IntegerField(default=0)),
                ('frags_strong', models.IntegerField(default=0)),
                ('acc_strong', models.DecimalField(max_digits=5, default=0, decimal_places=2)),
                ('shots_weak', models.IntegerField(default=0)),
                ('hits_weak', models.IntegerField(default=0)),
                ('dmg_weak', models.IntegerField(default=0)),
                ('frags_weak', models.IntegerField(default=0)),
                ('acc_weak', models.DecimalField(max_digits=5, default=0, decimal_places=2)),
                ('matchresult', models.ForeignKey(related_name='+', to='warmama.MatchResult')),
            ],
            options={
                'db_table': 'match_weapons',
            },
        ),
        migrations.CreateModel(
            name='Player',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, verbose_name='ID', serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('login', models.CharField(unique=True, max_length=64)),
                ('nickname', models.CharField(blank=True, max_length=64)),
                ('ip', models.CharField(blank=True, max_length=22)),
                ('ipv6', models.CharField(blank=True, max_length=54)),
                ('location', models.CharField(blank=True, db_index=True, max_length=2)),
                ('banned', models.BooleanField(default=False)),
            ],
            options={
                'db_table': 'players',
            },
        ),
        migrations.CreateModel(
            name='PlayerAchievement',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, verbose_name='ID', serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('achievement', models.ForeignKey(related_name='+', to='warmama.Achievement')),
                ('player', models.ForeignKey(related_name='achievements', to='warmama.Player')),
            ],
            options={
                'db_table': 'player_achievements',
            },
        ),
        migrations.CreateModel(
            name='PlayerSession',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, verbose_name='ID', serialize=False)),
                ('created', models.DateTimeField(db_index=True, auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('ip', models.CharField(blank=True, max_length=22)),
                ('ipv6', models.CharField(blank=True, max_length=54)),
                ('digest', models.CharField(default=uuid.uuid4, max_length=32)),
                ('ticket_id', models.IntegerField(null=True, blank=True, default=0)),
                ('ticket_expiration', models.DateTimeField(null=True, blank=True)),
                ('purgable', models.BooleanField(default=False)),
            ],
            options={
                'db_table': 'sessions_player',
            },
        ),
        migrations.CreateModel(
            name='PlayerStat',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, verbose_name='ID', serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(db_index=True, auto_now=True)),
                ('wins', models.IntegerField(default=0)),
                ('losses', models.IntegerField(default=0)),
                ('quits', models.IntegerField(default=0)),
                ('rating', models.DecimalField(max_digits=8, default=0, decimal_places=2)),
                ('deviation', models.DecimalField(max_digits=8, default=0, decimal_places=2)),
                ('gametype', models.ForeignKey(related_name='+', to='warmama.Gametype')),
                ('player', models.ForeignKey(related_name='stats', to='warmama.Player')),
            ],
            options={
                'db_table': 'player_stats',
            },
        ),
        migrations.CreateModel(
            name='PurgePlayer',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, verbose_name='ID', serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('player', models.OneToOneField(to='warmama.Player', related_name='+')),
            ],
            options={
                'db_table': 'purge_players',
            },
        ),
        migrations.CreateModel(
            name='RaceRun',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, verbose_name='ID', serialize=False)),
                ('created', models.DateTimeField(db_index=True, auto_now_add=True)),
                ('utctime', models.DateTimeField()),
                ('map', models.ForeignKey(related_name='+', to='warmama.Map')),
                ('player', models.ForeignKey(related_name='+', to='warmama.Player')),
            ],
            options={
                'db_table': 'race_runs',
            },
        ),
        migrations.CreateModel(
            name='RaceSector',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, verbose_name='ID', serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('sector', models.IntegerField()),
                ('time', models.IntegerField(db_index=True, default=0)),
                ('run', models.ForeignKey(related_name='sectors', to='warmama.RaceRun')),
            ],
            options={
                'db_table': 'race_sectors',
            },
        ),
        migrations.CreateModel(
            name='Server',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, verbose_name='ID', serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('login', models.CharField(unique=True, max_length=64)),
                ('regip', models.CharField(blank=True, db_index=True, max_length=22)),
                ('regipv6', models.CharField(blank=True, max_length=54)),
                ('hostname', models.CharField(blank=True, max_length=64)),
                ('ip', models.CharField(blank=True, db_index=True, max_length=22)),
                ('ipv6', models.CharField(blank=True, max_length=54)),
                ('location', models.CharField(blank=True, db_index=True, max_length=2)),
                ('banned', models.BooleanField(default=False)),
                ('demos_baseurl', models.CharField(blank=True, max_length=128)),
            ],
            options={
                'db_table': 'servers',
            },
        ),
        migrations.CreateModel(
            name='ServerSession',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, verbose_name='ID', serialize=False)),
                ('created', models.DateTimeField(db_index=True, auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('ip', models.CharField(blank=True, db_index=True, max_length=22)),
                ('ipv6', models.CharField(blank=True, max_length=54)),
                ('digest', models.CharField(default=uuid.uuid4, max_length=36)),
                ('port', models.IntegerField()),
                ('next_match_uuid', models.CharField(blank=True, unique=True, max_length=36)),
                ('user', models.OneToOneField(to='warmama.Server', related_name='session')),
            ],
            options={
                'db_table': 'sessions_server',
            },
        ),
        migrations.CreateModel(
            name='Weapon',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, verbose_name='ID', serialize=False)),
                ('name', models.CharField(unique=True, max_length=2)),
                ('fullname', models.CharField(max_length=16)),
            ],
            options={
                'db_table': 'weapons',
            },
        ),
        migrations.AddField(
            model_name='racerun',
            name='server',
            field=models.ForeignKey(related_name='+', to='warmama.Server'),
        ),
        migrations.AddField(
            model_name='purgeplayer',
            name='server_session',
            field=models.ForeignKey(blank=True, to='warmama.ServerSession', null=True, db_column='server_session', related_name='+', on_delete=django.db.models.deletion.SET_NULL),
        ),
        migrations.AddField(
            model_name='purgeplayer',
            name='session',
            field=models.OneToOneField(to='warmama.PlayerSession', related_name='purge_player'),
        ),
        migrations.AddField(
            model_name='playersession',
            name='server_session',
            field=models.ForeignKey(blank=True, to='warmama.ServerSession', null=True, db_column='server_session', related_name='+', on_delete=django.db.models.deletion.SET_NULL),
        ),
        migrations.AddField(
            model_name='playersession',
            name='ticket_server',
            field=models.ForeignKey(blank=True, to='warmama.ServerSession', null=True, db_column='ticket_server', related_name='+', on_delete=django.db.models.deletion.SET_NULL),
        ),
        migrations.AddField(
            model_name='playersession',
            name='user',
            field=models.OneToOneField(to='warmama.Player', related_name='session'),
        ),
        migrations.AddField(
            model_name='matchweapon',
            name='player',
            field=models.ForeignKey(related_name='+', to='warmama.Player'),
        ),
        migrations.AddField(
            model_name='matchweapon',
            name='weapon',
            field=models.ForeignKey(related_name='+', to='warmama.Weapon'),
        ),
        migrations.AddField(
            model_name='matchresult',
            name='server',
            field=models.ForeignKey(related_name='matches', to='warmama.Server'),
        ),
        migrations.AddField(
            model_name='matchresult',
            name='winner_player',
            field=models.ForeignKey(blank=True, to='warmama.MatchPlayer', null=True, related_name='+'),
        ),
        migrations.AddField(
            model_name='matchresult',
            name='winner_team',
            field=models.ForeignKey(blank=True, to='warmama.MatchTeam', null=True, related_name='+'),
        ),
        migrations.AddField(
            model_name='matchplayer',
            name='matchresult',
            field=models.ForeignKey(related_name='players', to='warmama.MatchResult'),
        ),
        migrations.AddField(
            model_name='matchplayer',
            name='matchteam',
            field=models.ForeignKey(blank=True, to='warmama.MatchTeam', null=True, related_name='players'),
        ),
        migrations.AddField(
            model_name='matchplayer',
            name='player',
            field=models.ForeignKey(related_name='matches', to='warmama.Player'),
        ),
        migrations.AddField(
            model_name='matchfrag',
            name='attacker',
            field=models.ForeignKey(related_name='+', to='warmama.Player'),
        ),
        migrations.AddField(
            model_name='matchfrag',
            name='matchresult',
            field=models.ForeignKey(related_name='frags', to='warmama.MatchResult'),
        ),
        migrations.AddField(
            model_name='matchfrag',
            name='victim',
            field=models.ForeignKey(related_name='+', to='warmama.Player'),
        ),
        migrations.AddField(
            model_name='matchfrag',
            name='weapon',
            field=models.ForeignKey(blank=True, to='warmama.Weapon', null=True),
        ),
        migrations.AddField(
            model_name='matchaward',
            name='matchresult',
            field=models.ForeignKey(related_name='+', to='warmama.MatchResult'),
        ),
        migrations.AddField(
            model_name='matchaward',
            name='player',
            field=models.ForeignKey(related_name='+', to='warmama.Player'),
        ),
        migrations.AlterUniqueTogether(
            name='racesector',
            unique_together=set([('run', 'sector')]),
        ),
        migrations.AlterUniqueTogether(
            name='playerstat',
            unique_together=set([('player', 'gametype')]),
        ),
        migrations.AlterUniqueTogether(
            name='matchteam',
            unique_together=set([('matchresult', 'name')]),
        ),
        migrations.AlterUniqueTogether(
            name='matchplayer',
            unique_together=set([('player', 'matchresult')]),
        ),
    ]



import binascii
import json
import logging
import zlib
from base64 import urlsafe_b64decode
from datetime import datetime
from django import forms
from django.core.exceptions import ValidationError
from django.utils import timezone
from rest_framework import serializers


logger = logging.getLogger(__name__)


##########
# Form Fields
##########


class IPAddressField(forms.GenericIPAddressField):
    """IPAddressField which strips port

    Splits the port from the value before interpreting it as an ip address.
    For ipv6 this takes the non-standard form `ip:port`, if port is specified
    empty groups must be explicitly written. (e.g.
    `2001:db8:85a3::::370:7334:port` not `2001:db8:85a3::370:7334:port`)
    """
    def to_python(self, value):
        try:
            value = value.strip()
        except AttributeError:
            raise ValidationError('Value must be a string', code='invalid')

        addr, port = value, None
        if '.' in value and ':' in value:
            # ipv4 or ipv6-packed ipv4
            port_start = value.rfind(':', value.rfind('.'))
            if port_start != -1:
                addr, port = value[:port_start], value[port_start + 1:]
        elif value.count(':') == 8:
            # ipv6 with port
            addr, port = value.rsplit(':', 1)

        addr = super(IPAddressField, self).to_python(addr)
        if port:
            try:
                port = int(port)
            except ValueError:
                raise ValidationError('Port must be an integer')
        else:
            port = None

        return addr, port

    def validate(self, value):
        addr, port = value
        super(IPAddressField, self).validate(addr)

    def run_validators(self, value):
        addr, port = value
        super(IPAddressField, self).run_validators(addr)


class GzipJsonField(forms.Field):
    """Field which expects urlsafe-base64 encoded gzipped data"""

    def to_python(self, value):
        # Value must be `Unicode` type on PY2
        try:
            value = value.encode('utf-8')
        except ValueError:
            raise ValidationError('Couldnt UTF-8 encode value', code='invalid')

        try:
            value = urlsafe_b64decode(value)
        except (TypeError, binascii.Error):
            raise ValidationError('Couldnt base64 decode value', code='invalid')

        try:
            value = zlib.decompress(value)
        except zlib.error:
            raise ValidationError('Couldnt zlib decompress value', code='invalid')

        try:
            value = value.decode('utf-8')
        except ValueError:
            raise ValidationError('Couldnt UTF-8 decode the decompressed value', code='invalid')

        try:
            value = json.loads(value)
        except (TypeError, ValueError):
            raise ValidationError('Couldnt parse value as JSON', code='invalid')

        return super(GzipJsonField, self).to_python(value)


##########
# Serializer Fields
##########


class TimestampField(serializers.Field):
    """Convert POSIX timestamps to/from TZ-aware datetime objects"""
    default_error_messages = {
        'invalid': 'Must be a TZ-aware datetime object',
    }

    def to_representation(self, obj):
        """Convert a TZ-aware datetime to its POSIX timestamp"""
        if timezone.is_naive(obj):
            self.fail('invalid')
        return (obj - datetime(1970, 1, 1, tzinfo=timezone.utc)).total_seconds()

    def to_internal_value(self, data):
        """Convert a POSIX timestamp to a TZ-aware datetime"""
        return datetime.utcfromtimestamp(data).replace(tzinfo=timezone.utc)



from django.conf.urls import url
from warmama import views

urlpatterns = [
    url(r'^slogin$', views.ServerLogin.as_view(), name='slogin'),
    url(r'^slogout$', views.ServerLogout.as_view(), name='slogout'),
    url(r'^scc$', views.ServerClientConnect.as_view(), name='scc'),
    url(r'^scd$', views.ServerClientDisconnect.as_view(), name='scd'),
    url(r'^smr$', views.ServerMatchReport.as_view(), name='smr'),
    url(r'^shb$', views.ServerHeartbeat.as_view(), name='shb'),
    url(r'^smuuid$', views.ServerMatchUUID.as_view(), name='smuuid'),
    url(r'^clogin$', views.ClientLogin.as_view(), name='clogin'),
    url(r'^clogout$', views.ClientLogout.as_view(), name='clogout'),
    url(r'^ccc$', views.ClientConnect.as_view(), name='ccc'),
    url(r'^chb$', views.ClientHeartbeat.as_view(), name='chb'),
    url(r'^auth$', views.ClientAuthenticate.as_view(), name='auth'),
]



import logging
import traceback
from django.conf import settings
from django.core import exceptions as djexceptions
from django.http import JsonResponse
from functools import wraps
from ipware.ip import get_ip
from ipware.utils import is_valid_ipv4, is_valid_ipv6
from warmama import exceptions as wmexceptions


logger = logging.getLogger(__name__)


def exception_response(data):
    """Decorator factory to return JSON response on exception

    Args:
        data (dict): Default JSON data to return.
    """
    def decorator(func):
        @wraps(func)
        def wrapped(*args, **kwargs):
            try:
                result = func(*args, **kwargs)
            except wmexceptions.WarmamaException as e:
                logger.info('WarmamaException: %s', e)
                return JsonResponse(data, status=e.status)
            except djexceptions.ObjectDoesNotExist as e:
                # Assume DNE means unauthorized
                logger.info('DoesNotExist: %s', e)
                return JsonResponse(data, status=403)
            except Exception as e:
                logger.error('Unexpected exception: %s', e)
                logger.error('Traceback: %s', traceback.format_exc())
                return JsonResponse(data, status=500)

            return result

        return wrapped
    return decorator


def request_ip(func):
    """View function decorator to add ipv4 and ipv6 arguments

    Calculates the ip address from the request and passes it to the function
    as named arguments `ipv4` and `ipv6`. Accepts private ip addresses if
    `DEBUG` is `True`.
    """
    @wraps(func)
    def wrapped(request, *args, **kwargs):
        # If either one is already passed, preserve it
        ipv4, ipv6 = kwargs.get('ipv4', ''), kwargs.get('ipv6', '')
        if ipv4 or ipv6:
            kwargs['ipv4'] = ipv4
            kwargs['ipv6'] = ipv6
            return func(request, *args, **kwargs)

        real_ip = not settings.DEBUG
        ipaddr = get_ip(request, real_ip_only=real_ip)
        if not ipaddr:
            raise wmexceptions.BadRequest('Couldnt read IP address for request')

        if is_valid_ipv4(ipaddr):
            kwargs['ipv4'], kwargs['ipv6'] = ipaddr, ''
        elif is_valid_ipv6(ipaddr):
            kwargs['ipv4'], kwargs['ipv6'] = '', ipaddr
        else:
            raise wmexceptions.BadRequest('Couldnt read IP address for request')

        return func(request, *args, **kwargs)

    return wrapped


def clean_post_args(form):
    """Factory for view function decorators to Form validate POST arguments

    Uses the `form` to validate `request.POST` data and passes it as a argument
    to the wrapped function.
    """
    def decorator(func):
        @wraps(func)
        def wrapped(request, *args, **kwargs):
            # If already passed, preserve it
            if kwargs.get('POST', None):
                return func(request, *args, **kwargs)

            formdata = form(request.POST)
            if not formdata.is_valid():
                logger.info('Invalid form data: %s', formdata.errors)
                raise wmexceptions.BadRequest(
                    'Invalid arguments: {0}'.format(formdata.errors)
                )
            kwargs['POST'] = formdata.cleaned_data
            return func(request, *args, **kwargs)

        return wrapped
    return decorator


def log_args(logger):
    """Factor for decorator to log function calls/arguments with level DEBUG"""
    def decorator(func):
        @wraps(func)
        def wrapped(*args, **kwargs):
            logger.debug('args: %s', args)
            logger.debug('kwargs: %s', kwargs)
            return func(*args, **kwargs)

        return wrapped
    return decorator
